@@ -35,11 +35,11 @@
 
 #include <cmath>
 #include <cstdlib>
-
-#include "src/base/platform/platform-posix.h"
+#include <optional>
 
 #include "src/base/lazy-instance.h"
 #include "src/base/macros.h"
+#include "src/base/platform/platform-posix.h"
 #include "src/base/platform/platform.h"
 #include "src/base/platform/time.h"
 #include "src/base/utils/random-number-generator.h"
@@ -55,7 +55,7 @@
 #if V8_OS_DARWIN
 #include <mach/mach.h>
 #include <malloc/malloc.h>
-#else
+#elif !V8_OS_ZOS
 #include <malloc.h>
 #endif
 
@@ -69,7 +69,7 @@
 #include <sys/resource.h>
 #endif
 
-#if !defined(_AIX) && !defined(V8_OS_FUCHSIA)
+#if !defined(_AIX) && !defined(V8_OS_FUCHSIA) && !V8_OS_ZOS
 #include <sys/syscall.h>
 #endif
 
@@ -99,9 +99,11 @@ namespace base {
 namespace {
 
 // 0 is never a valid thread id.
+#if V8_OS_ZOS
+const pthread_t kNoThread = {0};  // pthread_t is a struct on z/OS
+#else
 const pthread_t kNoThread = static_cast<pthread_t>(0);
-
-bool g_hard_abort = false;
+#endif
 
 const char* g_gc_fake_mmap = nullptr;
 
@@ -109,7 +111,7 @@ DEFINE_LAZY_LEAKY_OBJECT_GETTER(RandomNumberGenerator,
                                 GetPlatformRandomNumberGenerator)
 static LazyMutex rng_mutex = LAZY_MUTEX_INITIALIZER;
 
-#if !V8_OS_FUCHSIA
+#if !V8_OS_FUCHSIA && !V8_OS_ZOS
 #if V8_OS_DARWIN
 // kMmapFd is used to pass vm_alloc flags to tag the region with the user
 // defined tag 255 This helps identify V8-allocated regions in memory analysis
@@ -134,7 +136,8 @@ int GetFlagsForMemoryPermission(OS::MemoryPermission access,
                                 PageType page_type) {
   int flags = MAP_ANONYMOUS;
   flags |= (page_type == PageType::kShared) ? MAP_SHARED : MAP_PRIVATE;
-  if (access == OS::MemoryPermission::kNoAccess) {
+  if (access == OS::MemoryPermission::kNoAccess ||
+      access == OS::MemoryPermission::kNoAccessWillJitLater) {
 #if !V8_OS_AIX && !V8_OS_FREEBSD && !V8_OS_QNX
     flags |= MAP_NORESERVE;
 #endif  // !V8_OS_AIX && !V8_OS_FREEBSD && !V8_OS_QNX
@@ -147,7 +150,8 @@ int GetFlagsForMemoryPermission(OS::MemoryPermission access,
   // hardened runtime/memory protection is enabled, which is optional (via code
   // signing) on Intel-based Macs but mandatory on Apple silicon ones. See also
   // https://developer.apple.com/documentation/apple-silicon/porting-just-in-time-compilers-to-apple-silicon.
-  if (access == OS::MemoryPermission::kNoAccessWillJitLater) {
+  if (access == OS::MemoryPermission::kNoAccessWillJitLater ||
+      access == OS::MemoryPermission::kReadWriteExecute) {
     flags |= MAP_JIT;
   }
 #endif  // V8_OS_DARWIN
@@ -184,7 +188,7 @@ void* Allocate(void* hint, size_t size, OS::MemoryPermission access,
   return result;
 }
 
-#endif  // !V8_OS_FUCHSIA
+#endif  // !V8_OS_FUCHSIA && !V8_OS_ZOS
 
 }  // namespace
 
@@ -251,17 +255,20 @@ bool OS::ArmUsingHardFloat() {
 #endif  // def __arm__
 #endif
 
-void PosixInitializeCommon(bool hard_abort, const char* const gc_fake_mmap) {
-  g_hard_abort = hard_abort;
+void PosixInitializeCommon(AbortMode abort_mode,
+                           const char* const gc_fake_mmap) {
+  g_abort_mode = abort_mode;
   g_gc_fake_mmap = gc_fake_mmap;
 }
 
 #if !V8_OS_FUCHSIA
-void OS::Initialize(bool hard_abort, const char* const gc_fake_mmap) {
-  PosixInitializeCommon(hard_abort, gc_fake_mmap);
+void OS::Initialize(AbortMode abort_mode, const char* const gc_fake_mmap) {
+  PosixInitializeCommon(abort_mode, gc_fake_mmap);
 }
 #endif  // !V8_OS_FUCHSIA
 
+bool OS::IsHardwareEnforcedShadowStacksEnabled() { return false; }
+
 int OS::ActivationFrameAlignment() {
 #if V8_TARGET_ARCH_ARM
   // On EABI ARM targets this is required for fp correctness in the
@@ -370,9 +377,9 @@ void* OS::GetRandomMmapAddr() {
   // this address for RISC-V. https://github.com/v8-riscv/v8/issues/375
   raw_addr &= 0x3FFFF000;
 #elif V8_TARGET_ARCH_LOONG64
-  // 42 bits of virtual addressing. Truncate to 40 bits to allow kernel chance
-  // to fulfill request.
-  raw_addr &= uint64_t{0xFFFFFF0000};
+  // 40 or 47 bits of virtual addressing. Truncate to 38 bits to allow kernel
+  // chance to fulfill request.
+  raw_addr &= uint64_t{0x3FFFFF0000};
 #else
   raw_addr &= 0x3FFFF000;
 
@@ -404,6 +411,7 @@ void* OS::GetRandomMmapAddr() {
 
 // TODO(bbudge) Move Cygwin and Fuchsia stuff into platform-specific files.
 #if !V8_OS_CYGWIN && !V8_OS_FUCHSIA
+#if !V8_OS_ZOS
 // static
 void* OS::Allocate(void* hint, size_t size, size_t alignment,
                    MemoryPermission access) {
@@ -541,10 +549,8 @@ bool OS::RecommitPages(void* address, size_t size, MemoryPermission access) {
 #if defined(V8_OS_DARWIN)
   while (madvise(address, size, MADV_FREE_REUSE) == -1 && errno == EAGAIN) {
   }
-  return true;
-#else
-  return SetPermissions(address, size, access);
 #endif  // defined(V8_OS_DARWIN)
+  return true;
 }
 
 // static
@@ -605,12 +611,13 @@ bool OS::DecommitPages(void* address, size_t size) {
   return true;
 }
 #endif  // !defined(_AIX)
+#endif  // !V8_OS_ZOS
 
 // static
 bool OS::CanReserveAddressSpace() { return true; }
 
 // static
-Optional<AddressSpaceReservation> OS::CreateAddressSpaceReservation(
+std::optional<AddressSpaceReservation> OS::CreateAddressSpaceReservation(
     void* hint, size_t size, size_t alignment,
     MemoryPermission max_permission) {
   // On POSIX, address space reservations are backed by private memory mappings.
@@ -672,6 +679,7 @@ void OS::DestroySharedMemoryHandle(PlatformSharedMemoryHandle handle) {
 }
 #endif  // !defined(V8_OS_DARWIN)
 
+#if !V8_OS_ZOS
 // static
 bool OS::HasLazyCommits() {
 #if V8_OS_AIX || V8_OS_LINUX || V8_OS_DARWIN
@@ -681,6 +689,7 @@ bool OS::HasLazyCommits() {
   return false;
 #endif
 }
+#endif  // !V8_OS_ZOS
 #endif  // !V8_OS_CYGWIN && !V8_OS_FUCHSIA
 
 const char* OS::GetGCFakeMMapFile() {
@@ -694,8 +703,15 @@ void OS::Sleep(TimeDelta interval) {
 
 
 void OS::Abort() {
-  if (g_hard_abort) {
-    IMMEDIATE_CRASH();
+  switch (g_abort_mode) {
+    case AbortMode::kExitWithSuccessAndIgnoreDcheckFailures:
+      _exit(0);
+    case AbortMode::kExitWithFailureAndIgnoreDcheckFailures:
+      _exit(-1);
+    case AbortMode::kImmediateCrash:
+      IMMEDIATE_CRASH();
+    case AbortMode::kDefault:
+      break;
   }
   // Redirect to std abort to signal abnormal program termination.
   abort();
@@ -719,6 +735,8 @@ void OS::DebugBreak() {
   asm("int $3");
 #elif V8_HOST_ARCH_X64
   asm("int $3");
+#elif V8_OS_ZOS
+  asm(" dc x'0001'");
 #elif V8_HOST_ARCH_S390
   // Software breakpoint instruction is 0x0001
   asm volatile(".word 0x0001");
@@ -731,7 +749,7 @@ void OS::DebugBreak() {
 #endif
 }
 
-
+#if !V8_OS_ZOS
 class PosixMemoryMappedFile final : public OS::MemoryMappedFile {
  public:
   PosixMemoryMappedFile(FILE* file, void* memory, size_t size)
@@ -801,7 +819,7 @@ PosixMemoryMappedFile::~PosixMemoryMappedFile() {
   if (memory_) OS::Free(memory_, RoundUp(size_, OS::AllocatePageSize()));
   fclose(file_);
 }
-
+#endif  // !V8_OS_ZOS
 
 int OS::GetCurrentProcessId() {
   return static_cast<int>(getpid());
@@ -821,6 +839,8 @@ int OS::GetCurrentThreadId() {
   return static_cast<int>(zx_thread_self());
 #elif V8_OS_SOLARIS
   return static_cast<int>(pthread_self());
+#elif V8_OS_ZOS
+  return gettid();
 #else
   return static_cast<int>(reinterpret_cast<intptr_t>(pthread_self()));
 #endif
@@ -849,6 +869,28 @@ int OS::GetUserTime(uint32_t* secs, uint32_t* usecs) {
 }
 #endif
 
+int OS::GetPeakMemoryUsageKb() {
+#if defined(V8_OS_FUCHSIA)
+  // Fuchsia does not implement getrusage()
+  return -1;
+#elif defined(V8_OS_ZOS)
+  // TODO(v8:342445981): zos - rusage struct doesn't yet include ru_maxrss
+  return -1;
+#else
+  struct rusage usage;
+  if (getrusage(RUSAGE_SELF, &usage) < 0) return -1;
+
+#if defined(V8_OS_MACOS) || defined(V8_OS_IOS)
+  constexpr int KB = 1024;
+  // MacOS and iOS ru_maxrss count bytes
+  return static_cast<int>(usage.ru_maxrss / KB);
+#else
+  // Most other cases (at least Linux, IOS, return kilobytes)
+  return static_cast<int>(usage.ru_maxrss);
+#endif  // defined(V8_OS_MACOS) || defined(V8_OS_IOS)
+#endif  // defined(V8_OS_FUCHSIA)
+}
+
 double OS::TimeCurrentMillis() {
   return Time::Now().ToJsTime();
 }
@@ -945,6 +987,7 @@ void OS::PrintError(const char* format, ...) {
   va_start(args, format);
   VPrintError(format, args);
   va_end(args);
+  fflush(stderr);
 }
 
 
@@ -996,7 +1039,8 @@ void OS::StrNCpy(char* dest, int length, const char* src, size_t n) {
 
 #if !V8_OS_CYGWIN && !V8_OS_FUCHSIA
 
-Optional<AddressSpaceReservation> AddressSpaceReservation::CreateSubReservation(
+std::optional<AddressSpaceReservation>
+AddressSpaceReservation::CreateSubReservation(
     void* address, size_t size, OS::MemoryPermission max_permission) {
   DCHECK(Contains(address, size));
   DCHECK_EQ(0, size % OS::AllocatePageSize());
@@ -1030,6 +1074,8 @@ bool AddressSpaceReservation::Free(void* address, size_t size) {
   return OS::DecommitPages(address, size);
 }
 
+// z/OS specific implementation in platform-zos.cc.
+#if !defined(V8_OS_ZOS)
 // Darwin specific implementation in platform-darwin.cc.
 #if !defined(V8_OS_DARWIN)
 bool AddressSpaceReservation::AllocateShared(void* address, size_t size,
@@ -1049,6 +1095,7 @@ bool AddressSpaceReservation::FreeShared(void* address, size_t size) {
   return mmap(address, size, PROT_NONE, MAP_FIXED | MAP_ANONYMOUS | MAP_PRIVATE,
               -1, 0) == address;
 }
+#endif  // !V8_OS_ZOS
 
 bool AddressSpaceReservation::SetPermissions(void* address, size_t size,
                                              OS::MemoryPermission access) {
@@ -1147,7 +1194,7 @@ static void* ThreadEntry(void* arg) {
     case Thread::Priority::kDefault:
       break;
   }
-#elif V8_OS_LINUX
+#elif V8_OS_LINUX || V8_OS_ZOS
   switch (thread->priority()) {
     case Thread::Priority::kBestEffort:
       setpriority(PRIO_PROCESS, 0, 10);
@@ -1300,6 +1347,9 @@ bool MainThreadIsCurrentThread() {
 
 // static
 Stack::StackSlot Stack::ObtainCurrentThreadStackStart() {
+#if V8_OS_ZOS
+  return __get_stack_start();
+#else
   pthread_attr_t attr;
   int error = pthread_getattr_np(pthread_self(), &attr);
   if (error) {
@@ -1329,6 +1379,7 @@ Stack::StackSlot Stack::ObtainCurrentThreadStackStart() {
   }
 #endif  // !defined(V8_LIBC_GLIBC)
   return stack_start;
+#endif  // V8_OS_ZOS
 }
 
 #endif  // !defined(V8_OS_FREEBSD) && !defined(V8_OS_DARWIN) &&
