@@ -2,6 +2,8 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
+#include <optional>
+
 #include "src/builtins/builtins-utils-inl.h"
 #include "src/builtins/builtins.h"
 #include "src/heap/heap-inl.h"  // For ToBoolean. TODO(jkummerow): Drop.
@@ -28,12 +30,14 @@ bool IsValidCodePoint(Isolate* isolate, Handle<Object> value) {
     return false;
   }
 
-  if (Object::Number(*Object::ToInteger(isolate, value).ToHandleChecked()) !=
-      Object::Number(*value)) {
+  if (Object::NumberValue(
+          *Object::ToInteger(isolate, value).ToHandleChecked()) !=
+      Object::NumberValue(*value)) {
     return false;
   }
 
-  if (Object::Number(*value) < 0 || Object::Number(*value) > 0x10FFFF) {
+  if (Object::NumberValue(*value) < 0 ||
+      Object::NumberValue(*value) > 0x10FFFF) {
     return false;
   }
 
@@ -51,7 +55,7 @@ base::uc32 NextCodePoint(Isolate* isolate, BuiltinArguments args, int index) {
         MessageTemplate::kInvalidCodePoint, value));
     return kInvalidCodePoint;
   }
-  return DoubleToUint32(Object::Number(*value));
+  return DoubleToUint32(Object::NumberValue(*value));
 }
 
 }  // namespace
@@ -146,11 +150,11 @@ BUILTIN(StringPrototypeLocaleCompare) {
   Handle<String> str2;
   ASSIGN_RETURN_FAILURE_ON_EXCEPTION(
       isolate, str2, Object::ToString(isolate, args.atOrUndefined(isolate, 1)));
-  base::Optional<int> result = Intl::StringLocaleCompare(
+  std::optional<int> result = Intl::StringLocaleCompare(
       isolate, str1, str2, args.atOrUndefined(isolate, 2),
       args.atOrUndefined(isolate, 3), kMethod);
   if (!result.has_value()) {
-    DCHECK(isolate->has_pending_exception());
+    DCHECK(isolate->has_exception());
     return ReadOnlyRoots(isolate).exception();
   }
   return Smi::FromInt(result.value());
@@ -244,9 +248,9 @@ inline bool ToUpperOverflows(base::uc32 character) {
 }
 
 template <class Converter>
-V8_WARN_UNUSED_RESULT static Object ConvertCaseHelper(
-    Isolate* isolate, String string, SeqString result, int result_length,
-    unibrow::Mapping<Converter, 128>* mapping) {
+V8_WARN_UNUSED_RESULT static Tagged<Object> ConvertCaseHelper(
+    Isolate* isolate, Tagged<String> string, Tagged<SeqString> result,
+    int result_length, unibrow::Mapping<Converter, 128>* mapping) {
   DisallowGarbageCollection no_gc;
   // We try this twice, once with the assumption that the result is no longer
   // than the input and, if that assumption breaks, again with the exact
@@ -272,16 +276,16 @@ V8_WARN_UNUSED_RESULT static Object ConvertCaseHelper(
     int char_length = mapping->get(current, next, chars);
     if (char_length == 0) {
       // The case conversion of this character is the character itself.
-      result.Set(i, current);
+      result->Set(i, current);
       i++;
     } else if (char_length == 1 &&
                (ignore_overflow || !ToUpperOverflows(current))) {
       // Common case: converting the letter resulted in one character.
       DCHECK(static_cast<base::uc32>(chars[0]) != current);
-      result.Set(i, chars[0]);
+      result->Set(i, chars[0]);
       has_changed_character = true;
       i++;
-    } else if (result_length == string.length()) {
+    } else if (result_length == string->length()) {
       bool overflows = ToUpperOverflows(current);
       // We've assumed that the result would be as long as the
       // input but here is a character that converts to several
@@ -322,7 +326,7 @@ V8_WARN_UNUSED_RESULT static Object ConvertCaseHelper(
                                              : Smi::FromInt(current_length);
     } else {
       for (int j = 0; j < char_length; j++) {
-        result.Set(i, chars[j]);
+        result->Set(i, chars[j]);
         i++;
       }
       has_changed_character = true;
@@ -341,7 +345,7 @@ V8_WARN_UNUSED_RESULT static Object ConvertCaseHelper(
 }
 
 template <class Converter>
-V8_WARN_UNUSED_RESULT static Object ConvertCase(
+V8_WARN_UNUSED_RESULT static Tagged<Object> ConvertCase(
     Handle<String> s, Isolate* isolate,
     unibrow::Mapping<Converter, 128>* mapping) {
   s = String::Flatten(isolate, s);
@@ -379,7 +383,8 @@ V8_WARN_UNUSED_RESULT static Object ConvertCase(
     result = isolate->factory()->NewRawTwoByteString(length).ToHandleChecked();
   }
 
-  Object answer = ConvertCaseHelper(isolate, *s, *result, length, mapping);
+  Tagged<Object> answer =
+      ConvertCaseHelper(isolate, *s, *result, length, mapping);
   if (IsException(answer, isolate) || IsString(answer)) return answer;
 
   DCHECK(IsSmi(answer));
@@ -454,7 +459,7 @@ BUILTIN(StringRaw) {
   IncrementalStringBuilder result_builder(isolate);
   // Intentional spec violation: we ignore {length} values >= 2^32, because
   // assuming non-empty chunks they would generate too-long strings anyway.
-  const double raw_len_number = Object::Number(*raw_len);
+  const double raw_len_number = Object::NumberValue(*raw_len);
   const uint32_t length = raw_len_number > std::numeric_limits<uint32_t>::max()
                               ? std::numeric_limits<uint32_t>::max()
                               : static_cast<uint32_t>(raw_len_number);
