@@ -4,14 +4,17 @@
 
 #include <fstream>
 
-#include "src/builtins/builtins.h"
+#include "src/builtins/builtins-inl.h"
 #include "src/builtins/profile-data-reader.h"
 #include "src/codegen/assembler-inl.h"
 #include "src/codegen/interface-descriptors.h"
 #include "src/codegen/macro-assembler-inl.h"
 #include "src/codegen/macro-assembler.h"
 #include "src/codegen/reloc-info-inl.h"
+#include "src/common/globals.h"
 #include "src/compiler/code-assembler.h"
+#include "src/compiler/pipeline.h"
+#include "src/compiler/turboshaft/phase.h"
 #include "src/execution/isolate.h"
 #include "src/handles/handles-inl.h"
 #include "src/heap/heap-inl.h"
@@ -23,6 +26,10 @@
 #include "src/objects/shared-function-info.h"
 #include "src/objects/smi.h"
 
+#if V8_ENABLE_WEBASSEMBLY
+#include "src/wasm/wasm-builtin-list.h"
+#endif
+
 namespace v8 {
 namespace internal {
 
@@ -41,6 +48,14 @@ AssemblerOptions BuiltinAssemblerOptions(Isolate* isolate, Builtin builtin) {
   CHECK(!options.isolate_independent_code);
   CHECK(!options.collect_win64_unwind_info);
 
+#if V8_ENABLE_WEBASSEMBLY
+  if (wasm::BuiltinLookup::IsWasmBuiltinId(builtin) ||
+      builtin == Builtin::kJSToWasmWrapper ||
+      builtin == Builtin::kJSToWasmHandleReturns ||
+      builtin == Builtin::kWasmToJsWrapperCSA) {
+    options.is_wasm = true;
+  }
+#endif
   if (!isolate->IsGeneratingEmbeddedBuiltins()) {
     return options;
   }
@@ -78,6 +93,9 @@ AssemblerOptions BuiltinAssemblerOptions(Isolate* isolate, Builtin builtin) {
 
 using MacroAssemblerGenerator = void (*)(MacroAssembler*);
 using CodeAssemblerGenerator = void (*)(compiler::CodeAssemblerState*);
+using TurboshaftAssemblerGenerator =
+    void (*)(compiler::turboshaft::PipelineData*, Isolate*,
+             compiler::turboshaft::Graph&, Zone*);
 
 Handle<Code> BuildPlaceholder(Isolate* isolate, Builtin builtin) {
   HandleScope scope(isolate);
@@ -101,9 +119,9 @@ Handle<Code> BuildPlaceholder(Isolate* isolate, Builtin builtin) {
   return scope.CloseAndEscape(code);
 }
 
-Tagged<Code> BuildWithMacroAssembler(Isolate* isolate, Builtin builtin,
-                                     MacroAssemblerGenerator generator,
-                                     const char* s_name) {
+V8_NOINLINE Tagged<Code> BuildWithMacroAssembler(
+    Isolate* isolate, Builtin builtin, MacroAssemblerGenerator generator,
+    const char* s_name) {
   HandleScope scope(isolate);
   uint8_t buffer[kBufferSize];
 
@@ -125,9 +143,17 @@ Tagged<Code> BuildWithMacroAssembler(Isolate* isolate, Builtin builtin,
     handler_table_offset = HandlerTable::EmitReturnTableStart(&masm);
     HandlerTable::EmitReturnEntry(
         &masm, 0, isolate->builtins()->js_entry_handler_offset());
+#if V8_ENABLE_DRUMBRAKE
+  } else if (builtin == Builtin::kWasmInterpreterCWasmEntry) {
+    handler_table_offset = HandlerTable::EmitReturnTableStart(&masm);
+    HandlerTable::EmitReturnEntry(
+        &masm, 0,
+        isolate->builtins()->cwasm_interpreter_entry_handler_offset());
+#endif  // V8_ENABLE_DRUMBRAKE
   }
-#if V8_ENABLE_WEBASSEMBLY && (V8_TARGET_ARCH_X64 || V8_TARGET_ARCH_ARM64)
-  // TODO(v8:12191): Enable on all platforms once the builtin has been ported.
+#if V8_ENABLE_WEBASSEMBLY &&                                              \
+    (V8_TARGET_ARCH_X64 || V8_TARGET_ARCH_ARM64 || V8_TARGET_ARCH_IA32 || \
+     V8_TARGET_ARCH_ARM || V8_TARGET_ARCH_RISCV64 || V8_TARGET_ARCH_LOONG64)
   if (builtin == Builtin::kWasmReturnPromiseOnSuspendAsm) {
     handler_table_offset = HandlerTable::EmitReturnTableStart(&masm);
     HandlerTable::EmitReturnEntry(
@@ -139,10 +165,11 @@ Tagged<Code> BuildWithMacroAssembler(Isolate* isolate, Builtin builtin,
   masm.GetCode(isolate->main_thread_local_isolate(), &desc,
                MacroAssembler::kNoSafepointTable, handler_table_offset);
 
-  Handle<Code> code = Factory::CodeBuilder(isolate, desc, CodeKind::BUILTIN)
-                          .set_self_reference(masm.CodeObject())
-                          .set_builtin(builtin)
-                          .Build();
+  DirectHandle<Code> code =
+      Factory::CodeBuilder(isolate, desc, CodeKind::BUILTIN)
+          .set_self_reference(masm.CodeObject())
+          .set_builtin(builtin)
+          .Build();
 #if defined(V8_OS_WIN64)
   isolate->SetBuiltinUnwindData(builtin, masm.GetUnwindInfo());
 #endif  // V8_OS_WIN64
@@ -161,33 +188,103 @@ Tagged<Code> BuildAdaptor(Isolate* isolate, Builtin builtin,
   Builtins::Generate_Adaptor(&masm, builtin_address);
   CodeDesc desc;
   masm.GetCode(isolate, &desc);
-  Handle<Code> code = Factory::CodeBuilder(isolate, desc, CodeKind::BUILTIN)
-                          .set_self_reference(masm.CodeObject())
-                          .set_builtin(builtin)
-                          .Build();
+  DirectHandle<Code> code =
+      Factory::CodeBuilder(isolate, desc, CodeKind::BUILTIN)
+          .set_self_reference(masm.CodeObject())
+          .set_builtin(builtin)
+          .Build();
   return *code;
 }
 
+inline constexpr char kTempZoneName[] = "temp-zone";
+inline constexpr char kBuiltinCompilationZoneName[] =
+    "builtin-compilation-zone";
+
+Tagged<Code> BuildWithTurboshaftAssemblerImpl(
+    Isolate* isolate, Builtin builtin, TurboshaftAssemblerGenerator generator,
+    std::function<compiler::CallDescriptor*(Zone*)> call_descriptor_builder,
+    const char* name) {
+  HandleScope scope(isolate);
+  using namespace compiler::turboshaft;  // NOLINT(build/namespaces)
+
+  compiler::ZoneStats zone_stats(isolate->allocator());
+  ZoneWithName<kBuiltinCompilationZoneName> zone(&zone_stats,
+                                                 kBuiltinCompilationZoneName);
+  OptimizedCompilationInfo info(base::CStrVector(name), zone, CodeKind::BUILTIN,
+                                builtin);
+  compiler::CallDescriptor* call_descriptor = call_descriptor_builder(zone);
+
+  PipelineData data(&zone_stats, TurboshaftPipelineKind::kTSABuiltin, isolate,
+                    &info, BuiltinAssemblerOptions(isolate, builtin));
+  data.InitializeBuiltinComponent(call_descriptor);
+  data.InitializeGraphComponent(nullptr);
+  ZoneWithName<kTempZoneName> temp_zone(&zone_stats, kTempZoneName);
+  generator(&data, isolate, data.graph(), temp_zone);
+
+  DirectHandle<Code> code =
+      compiler::Pipeline::GenerateCodeForTurboshaftBuiltin(
+          &data, call_descriptor, builtin, name,
+          ProfileDataFromFile::TryRead(name))
+          .ToHandleChecked();
+  return *code;
+}
+
+// Builder for builtins implemented in Turboshaft with JS linkage.
+V8_NOINLINE Tagged<Code> BuildWithTurboshaftAssemblerJS(
+    Isolate* isolate, Builtin builtin, TurboshaftAssemblerGenerator generator,
+    int argc, const char* name) {
+  return BuildWithTurboshaftAssemblerImpl(
+      isolate, builtin, generator,
+      [argc](Zone* zone) {
+        return compiler::Linkage::GetJSCallDescriptor(
+            zone, false, argc, compiler::CallDescriptor::kCanUseRoots);
+      },
+      name);
+}
+
 // Builder for builtins implemented in TurboFan with JS linkage.
-Tagged<Code> BuildWithCodeStubAssemblerJS(Isolate* isolate, Builtin builtin,
-                                          CodeAssemblerGenerator generator,
-                                          int argc, const char* name) {
+V8_NOINLINE Tagged<Code> BuildWithCodeStubAssemblerJS(
+    Isolate* isolate, Builtin builtin, CodeAssemblerGenerator generator,
+    int argc, const char* name) {
+  // TODO(nicohartmann): Remove this once `BuildWithTurboshaftAssemblerCS` has
+  // an actual use.
+  USE(&BuildWithTurboshaftAssemblerJS);
   HandleScope scope(isolate);
 
   Zone zone(isolate->allocator(), ZONE_NAME, kCompressGraphZone);
   compiler::CodeAssemblerState state(isolate, &zone, argc, CodeKind::BUILTIN,
                                      name, builtin);
   generator(&state);
-  Handle<Code> code = compiler::CodeAssembler::GenerateCode(
+  DirectHandle<Code> code = compiler::CodeAssembler::GenerateCode(
       &state, BuiltinAssemblerOptions(isolate, builtin),
       ProfileDataFromFile::TryRead(name));
   return *code;
 }
 
+// Builder for builtins implemented in Turboshaft with CallStub linkage.
+V8_NOINLINE Tagged<Code> BuildWithTurboshaftAssemblerCS(
+    Isolate* isolate, Builtin builtin, TurboshaftAssemblerGenerator generator,
+    CallDescriptors::Key interface_descriptor, const char* name) {
+  return BuildWithTurboshaftAssemblerImpl(
+      isolate, builtin, generator,
+      [interface_descriptor](Zone* zone) {
+        CallInterfaceDescriptor descriptor(interface_descriptor);
+        DCHECK_LE(0, descriptor.GetRegisterParameterCount());
+        return compiler::Linkage::GetStubCallDescriptor(
+            zone, descriptor, descriptor.GetStackParameterCount(),
+            compiler::CallDescriptor::kNoFlags,
+            compiler::Operator::kNoProperties);
+      },
+      name);
+}
+
 // Builder for builtins implemented in TurboFan with CallStub linkage.
-Tagged<Code> BuildWithCodeStubAssemblerCS(
+V8_NOINLINE Tagged<Code> BuildWithCodeStubAssemblerCS(
     Isolate* isolate, Builtin builtin, CodeAssemblerGenerator generator,
     CallDescriptors::Key interface_descriptor, const char* name) {
+  // TODO(nicohartmann): Remove this once `BuildWithTurboshaftAssemblerCS` has
+  // an actual use.
+  USE(&BuildWithTurboshaftAssemblerCS);
   HandleScope scope(isolate);
   Zone zone(isolate->allocator(), ZONE_NAME, kCompressGraphZone);
   // The interface descriptor with given key must be initialized at this point
@@ -198,7 +295,7 @@ Tagged<Code> BuildWithCodeStubAssemblerCS(
   compiler::CodeAssemblerState state(isolate, &zone, descriptor,
                                      CodeKind::BUILTIN, name, builtin);
   generator(&state);
-  Handle<Code> code = compiler::CodeAssembler::GenerateCode(
+  DirectHandle<Code> code = compiler::CodeAssembler::GenerateCode(
       &state, BuiltinAssemblerOptions(isolate, builtin),
       ProfileDataFromFile::TryRead(name));
   return *code;
@@ -222,7 +319,7 @@ void SetupIsolateDelegate::PopulateWithPlaceholders(Isolate* isolate) {
   HandleScope scope(isolate);
   for (Builtin builtin = Builtins::kFirst; builtin <= Builtins::kLast;
        ++builtin) {
-    Handle<Code> placeholder = BuildPlaceholder(isolate, builtin);
+    DirectHandle<Code> placeholder = BuildPlaceholder(isolate, builtin);
     AddBuiltin(builtins, builtin, *placeholder);
   }
 }
@@ -242,10 +339,14 @@ void SetupIsolateDelegate::ReplacePlaceholders(Isolate* isolate) {
        ++builtin) {
     Tagged<Code> code = builtins->code(builtin);
     Tagged<InstructionStream> istream = code->instruction_stream();
-    CodePageMemoryModificationScope code_modification_scope(istream);
+    WritableJitAllocation jit_allocation = ThreadIsolation::LookupJitAllocation(
+        istream.address(), istream->Size(),
+        ThreadIsolation::JitAllocationType::kInstructionStream);
     bool flush_icache = false;
-    for (RelocIterator it(code, kRelocMask); !it.done(); it.next()) {
-      RelocInfo* rinfo = it.rinfo();
+    for (WritableRelocIterator it(jit_allocation, istream,
+                                  code->constant_pool(), kRelocMask);
+         !it.done(); it.next()) {
+      WritableRelocInfo* rinfo = it.rinfo();
       if (RelocInfo::IsCodeTargetMode(rinfo->rmode())) {
         Tagged<Code> target_code =
             Code::FromTargetAddress(rinfo->target_address());
@@ -260,7 +361,7 @@ void SetupIsolateDelegate::ReplacePlaceholders(Isolate* isolate) {
         DCHECK(RelocInfo::IsEmbeddedObjectMode(rinfo->rmode()));
         Tagged<Object> object = rinfo->target_object(cage_base);
         if (!IsCode(object, cage_base)) continue;
-        Tagged<Code> target = Code::cast(object);
+        Tagged<Code> target = Cast<Code>(object);
         if (!target->is_builtin()) continue;
         Tagged<Code> new_target = builtins->code(target->builtin_id());
         rinfo->set_target_object(istream, new_target, UPDATE_WRITE_BARRIER,
@@ -277,11 +378,11 @@ void SetupIsolateDelegate::ReplacePlaceholders(Isolate* isolate) {
 
 namespace {
 
-Tagged<Code> GenerateBytecodeHandler(Isolate* isolate, Builtin builtin,
-                                     interpreter::OperandScale operand_scale,
-                                     interpreter::Bytecode bytecode) {
+V8_NOINLINE Tagged<Code> GenerateBytecodeHandler(
+    Isolate* isolate, Builtin builtin, interpreter::OperandScale operand_scale,
+    interpreter::Bytecode bytecode) {
   DCHECK(interpreter::Bytecodes::BytecodeHasHandler(bytecode, operand_scale));
-  Handle<Code> code = interpreter::GenerateBytecodeHandler(
+  DirectHandle<Code> code = interpreter::GenerateBytecodeHandler(
       isolate, Builtins::name(builtin), bytecode, operand_scale, builtin,
       BuiltinAssemblerOptions(isolate, builtin));
   return *code;
@@ -312,12 +413,26 @@ void SetupIsolateDelegate::SetupBuiltinsInternal(Isolate* isolate) {
   AddBuiltin(builtins, Builtin::k##Name, code);              \
   index++;
 
+#define BUILD_TSJ(Name, Argc, ...)                                         \
+  code = BuildWithTurboshaftAssemblerJS(                                   \
+      isolate, Builtin::k##Name, &Builtins::Generate_##Name, Argc, #Name); \
+  AddBuiltin(builtins, Builtin::k##Name, code);                            \
+  index++;
+
 #define BUILD_TFJ(Name, Argc, ...)                                         \
   code = BuildWithCodeStubAssemblerJS(                                     \
       isolate, Builtin::k##Name, &Builtins::Generate_##Name, Argc, #Name); \
   AddBuiltin(builtins, Builtin::k##Name, code);                            \
   index++;
 
+#define BUILD_TSC(Name, InterfaceDescriptor)                      \
+  /* Return size is from the provided CallInterfaceDescriptor. */ \
+  code = BuildWithTurboshaftAssemblerCS(                          \
+      isolate, Builtin::k##Name, &Builtins::Generate_##Name,      \
+      CallDescriptors::InterfaceDescriptor, #Name);               \
+  AddBuiltin(builtins, Builtin::k##Name, code);                   \
+  index++;
+
 #define BUILD_TFC(Name, InterfaceDescriptor)                      \
   /* Return size is from the provided CallInterfaceDescriptor. */ \
   code = BuildWithCodeStubAssemblerCS(                            \
@@ -354,11 +469,13 @@ void SetupIsolateDelegate::SetupBuiltinsInternal(Isolate* isolate) {
   AddBuiltin(builtins, Builtin::k##Name, code);                     \
   index++;
 
-  BUILTIN_LIST(BUILD_CPP, BUILD_TFJ, BUILD_TFC, BUILD_TFS, BUILD_TFH, BUILD_BCH,
-               BUILD_ASM);
+  BUILTIN_LIST(BUILD_CPP, BUILD_TSJ, BUILD_TFJ, BUILD_TSC, BUILD_TFC, BUILD_TFS,
+               BUILD_TFH, BUILD_BCH, BUILD_ASM);
 
 #undef BUILD_CPP
+#undef BUILD_TSJ
 #undef BUILD_TFJ
+#undef BUILD_TSC
 #undef BUILD_TFC
 #undef BUILD_TFS
 #undef BUILD_TFH
