@@ -6,10 +6,10 @@
 
 #include <atomic>
 #include <iomanip>
+#include <optional>
 #include <sstream>
 
 #include "src/base/logging.h"
-#include "src/base/optional.h"
 #include "src/base/platform/mutex.h"
 #include "src/codegen/bailout-reason.h"
 #include "src/codegen/machine-type.h"
@@ -17,11 +17,13 @@
 #include "src/compiler/backend/instruction-selector.h"
 #include "src/compiler/frame-states.h"
 #include "src/compiler/graph-visualizer.h"
+#include "src/compiler/js-heap-broker.h"
 #include "src/compiler/machine-operator.h"
 #include "src/compiler/turboshaft/deopt-data.h"
 #include "src/compiler/turboshaft/graph.h"
 #include "src/handles/handles-inl.h"
 #include "src/handles/maybe-handles-inl.h"
+#include "src/objects/code-inl.h"
 
 #ifdef DEBUG
 // For InWritableSharedSpace
@@ -36,14 +38,69 @@ std::ostream& operator<<(std::ostream& os, AbortReason reason) {
 
 namespace v8::internal::compiler::turboshaft {
 
-void Print(const Operation& op) { std::cout << op << "\n"; }
+void Operation::Print() const { std::cout << *this << "\n"; }
 
 Zone* get_zone(Graph* graph) { return graph->graph_zone(); }
 
+std::optional<Builtin> TryGetBuiltinId(const ConstantOp* target,
+                                       JSHeapBroker* broker) {
+  if (!target) return std::nullopt;
+  if (target->kind != ConstantOp::Kind::kHeapObject) return std::nullopt;
+  // TODO(nicohartmann@): For builtin compilation we don't have a broker. We
+  // could try to access the heap directly instead.
+  if (broker == nullptr) return std::nullopt;
+  UnparkedScopeIfNeeded scope(broker);
+  AllowHandleDereference allow_handle_dereference;
+  HeapObjectRef ref = MakeRef(broker, target->handle());
+  if (ref.IsCode()) {
+    CodeRef code = ref.AsCode();
+    if (code.object()->is_builtin()) {
+      return code.object()->builtin_id();
+    }
+  }
+  return std::nullopt;
+}
+
+bool CallOp::IsStackCheck(const Graph& graph, JSHeapBroker* broker,
+                          StackCheckKind kind) const {
+  auto builtin_id =
+      TryGetBuiltinId(graph.Get(callee()).TryCast<ConstantOp>(), broker);
+  if (!builtin_id.has_value()) return false;
+  if (*builtin_id != Builtin::kCEntry_Return1_ArgvOnStack_NoBuiltinExit) {
+    return false;
+  }
+  DCHECK_GE(input_count, 4);
+  Runtime::FunctionId builtin = GetBuiltinForStackCheckKind(kind);
+  auto is_this_builtin = [&](int input_index) {
+    if (const ConstantOp* real_callee =
+            graph.Get(input(input_index)).TryCast<ConstantOp>();
+        real_callee != nullptr &&
+        real_callee->kind == ConstantOp::Kind::kExternal &&
+        real_callee->external_reference() ==
+            ExternalReference::Create(builtin)) {
+      return true;
+    }
+    return false;
+  };
+  // The function called by `CEntry_Return1_ArgvOnStack_NoBuiltinExit` is the
+  // 3rd or the 4th argument of the CallOp (depending on the stack check kind),
+  // so we check both of them.
+  return is_this_builtin(2) || is_this_builtin(3);
+}
+
+void CallOp::PrintOptions(std::ostream& os) const {
+  os << '[' << *descriptor->descriptor << ']';
+}
+
+void TailCallOp::PrintOptions(std::ostream& os) const {
+  os << '[' << *descriptor->descriptor << ']';
+}
+
+#if DEBUG
 bool ValidOpInputRep(
     const Graph& graph, OpIndex input,
     std::initializer_list<RegisterRepresentation> expected_reps,
-    base::Optional<size_t> projection_index) {
+    std::optional<size_t> projection_index) {
   base::Vector<const RegisterRepresentation> input_reps =
       graph.Get(input).outputs_rep();
   RegisterRepresentation input_rep;
@@ -68,7 +125,8 @@ bool ValidOpInputRep(
     return false;
   }
   for (RegisterRepresentation expected_rep : expected_reps) {
-    if (input_rep.AllowImplicitRepresentationChangeTo(expected_rep)) {
+    if (input_rep.AllowImplicitRepresentationChangeTo(
+            expected_rep, graph.IsCreatedFromTurbofan())) {
       return true;
     }
   }
@@ -82,9 +140,10 @@ bool ValidOpInputRep(
 
 bool ValidOpInputRep(const Graph& graph, OpIndex input,
                      RegisterRepresentation expected_rep,
-                     base::Optional<size_t> projection_index) {
+                     std::optional<size_t> projection_index) {
   return ValidOpInputRep(graph, input, {expected_rep}, projection_index);
 }
+#endif  // DEBUG
 
 const char* OpcodeName(Opcode opcode) {
 #define OPCODE_NAME(Name) #Name,
@@ -102,6 +161,26 @@ std::ostream& operator<<(std::ostream& os, OperationPrintStyle styled_op) {
   return os;
 }
 
+std::ostream& operator<<(std::ostream& os, GenericBinopOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(Name)              \
+  case GenericBinopOp::Kind::k##Name: \
+    return os << #Name;
+    GENERIC_BINOP_LIST(PRINT_KIND)
+#undef PRINT_KIND
+  }
+}
+
+std::ostream& operator<<(std::ostream& os, GenericUnopOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(Name)             \
+  case GenericUnopOp::Kind::k##Name: \
+    return os << #Name;
+    GENERIC_UNOP_LIST(PRINT_KIND)
+#undef PRINT_KIND
+  }
+}
+
 std::ostream& operator<<(std::ostream& os, WordUnaryOp::Kind kind) {
   switch (kind) {
     case WordUnaryOp::Kind::kReverseBytes:
@@ -119,6 +198,13 @@ std::ostream& operator<<(std::ostream& os, WordUnaryOp::Kind kind) {
   }
 }
 
+std::ostream& operator<<(std::ostream& os, OverflowCheckedUnaryOp::Kind kind) {
+  switch (kind) {
+    case OverflowCheckedUnaryOp::Kind::kAbs:
+      return os << "kAbs";
+  }
+}
+
 std::ostream& operator<<(std::ostream& os, FloatUnaryOp::Kind kind) {
   switch (kind) {
     case FloatUnaryOp::Kind::kAbs:
@@ -248,6 +334,8 @@ std::ostream& operator<<(std::ostream& os, ShiftOp::Kind kind) {
 
 std::ostream& operator<<(std::ostream& os, ComparisonOp::Kind kind) {
   switch (kind) {
+    case ComparisonOp::Kind::kEqual:
+      return os << "Equal";
     case ComparisonOp::Kind::kSignedLessThan:
       return os << "SignedLessThan";
     case ComparisonOp::Kind::kSignedLessThanOrEqual:
@@ -300,6 +388,8 @@ std::ostream& operator<<(std::ostream& os, ChangeOrDeoptOp::Kind kind) {
       return os << "Uint64ToInt64";
     case ChangeOrDeoptOp::Kind::kFloat64ToInt32:
       return os << "Float64ToInt32";
+    case ChangeOrDeoptOp::Kind::kFloat64ToUint32:
+      return os << "Float64ToUint32";
     case ChangeOrDeoptOp::Kind::kFloat64ToInt64:
       return os << "Float64ToInt64";
     case ChangeOrDeoptOp::Kind::kFloat64NotHole:
@@ -316,6 +406,19 @@ std::ostream& operator<<(std::ostream& os, TryChangeOp::Kind kind) {
   }
 }
 
+std::ostream& operator<<(std::ostream& os, TaggedBitcastOp::Kind kind) {
+  switch (kind) {
+    case TaggedBitcastOp::Kind::kSmi:
+      return os << "Smi";
+    case TaggedBitcastOp::Kind::kHeapObject:
+      return os << "HeapObject";
+    case TaggedBitcastOp::Kind::kTagAndSmiBits:
+      return os << "TagAndSmiBits";
+    case TaggedBitcastOp::Kind::kAny:
+      return os << "Any";
+  }
+}
+
 std::ostream& operator<<(std::ostream& os, ChangeOp::Assumption assumption) {
   switch (assumption) {
     case ChangeOp::Assumption::kNoAssumption:
@@ -355,25 +458,25 @@ std::ostream& operator<<(std::ostream& os, AtomicRMWOp::BinOp bin_op) {
   }
 }
 
-std::ostream& operator<<(std::ostream& os, AtomicWord32PairOp::OpKind bin_op) {
+std::ostream& operator<<(std::ostream& os, AtomicWord32PairOp::Kind bin_op) {
   switch (bin_op) {
-    case AtomicWord32PairOp::OpKind::kAdd:
+    case AtomicWord32PairOp::Kind::kAdd:
       return os << "add";
-    case AtomicWord32PairOp::OpKind::kSub:
+    case AtomicWord32PairOp::Kind::kSub:
       return os << "sub";
-    case AtomicWord32PairOp::OpKind::kAnd:
+    case AtomicWord32PairOp::Kind::kAnd:
       return os << "and";
-    case AtomicWord32PairOp::OpKind::kOr:
+    case AtomicWord32PairOp::Kind::kOr:
       return os << "or";
-    case AtomicWord32PairOp::OpKind::kXor:
+    case AtomicWord32PairOp::Kind::kXor:
       return os << "xor";
-    case AtomicWord32PairOp::OpKind::kExchange:
+    case AtomicWord32PairOp::Kind::kExchange:
       return os << "exchange";
-    case AtomicWord32PairOp::OpKind::kCompareExchange:
+    case AtomicWord32PairOp::Kind::kCompareExchange:
       return os << "compare-exchange";
-    case AtomicWord32PairOp::OpKind::kLoad:
+    case AtomicWord32PairOp::Kind::kLoad:
       return os << "load";
-    case AtomicWord32PairOp::OpKind::kStore:
+    case AtomicWord32PairOp::Kind::kStore:
       return os << "store";
   }
 }
@@ -413,7 +516,7 @@ void Operation::PrintOptions(std::ostream& os) const {
 }
 
 void ConstantOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kWord32:
       os << "word32: " << static_cast<int32_t>(storage.integral);
@@ -421,17 +524,30 @@ void ConstantOp::PrintOptions(std::ostream& os) const {
     case Kind::kWord64:
       os << "word64: " << static_cast<int64_t>(storage.integral);
       break;
+    case Kind::kSmi:
+      os << "smi: " << smi();
+      break;
     case Kind::kNumber:
-      os << "number: " << number();
+      os << "number: " << number().get_scalar();
       break;
     case Kind::kTaggedIndex:
       os << "tagged index: " << tagged_index();
       break;
     case Kind::kFloat64:
-      os << "float64: " << float64();
+      os << "float64: " << float64().get_scalar();
+      if (float64().is_hole_nan()) {
+        os << " (hole nan: 0x" << std::hex << float64().get_bits() << std::dec
+           << ')';
+      } else if (float64().is_nan()) {
+        os << " (0x" << std::hex << float64().get_bits() << std::dec << ')';
+      }
       break;
     case Kind::kFloat32:
-      os << "float32: " << float32();
+      os << "float32: " << float32().get_scalar();
+      if (float32().is_nan()) {
+        os << " (0x" << std::hex << base::bit_cast<uint32_t>(storage.float32)
+           << std::dec << ')';
+      }
       break;
     case Kind::kExternal:
       os << "external: " << external_reference();
@@ -442,6 +558,9 @@ void ConstantOp::PrintOptions(std::ostream& os) const {
     case Kind::kCompressedHeapObject:
       os << "compressed heap object: " << JSONEscaped(handle());
       break;
+    case Kind::kTrustedHeapObject:
+      os << "trusted heap object: " << JSONEscaped(handle());
+      break;
     case Kind::kRelocatableWasmCall:
       os << "relocatable wasm call: 0x"
          << reinterpret_cast<void*>(storage.integral);
@@ -450,14 +569,29 @@ void ConstantOp::PrintOptions(std::ostream& os) const {
       os << "relocatable wasm stub call: 0x"
          << reinterpret_cast<void*>(storage.integral);
       break;
+    case Kind::kRelocatableWasmCanonicalSignatureId:
+      os << "relocatable wasm canonical signature ID: "
+         << static_cast<int32_t>(storage.integral);
+      break;
   }
-  os << "]";
+  os << ']';
 }
 
 void ParameterOp::PrintOptions(std::ostream& os) const {
-  os << "[" << parameter_index;
+  os << '[' << parameter_index;
   if (debug_name) os << ", " << debug_name;
-  os << "]";
+  os << ']';
+}
+
+MachineType LoadOp::machine_type() const {
+  if (result_rep == RegisterRepresentation::Compressed()) {
+    if (loaded_rep == MemoryRepresentation::AnyTagged()) {
+      return MachineType::AnyCompressed();
+    } else if (loaded_rep == MemoryRepresentation::TaggedPointer()) {
+      return MachineType::CompressedPointer();
+    }
+  }
+  return loaded_rep.ToMachineType();
 }
 
 void LoadOp::PrintInputs(std::ostream& os,
@@ -469,66 +603,70 @@ void LoadOp::PrintInputs(std::ostream& os,
     os << " + " << offset;
   }
   if (index().valid()) {
-    os << " + " << op_index_prefix << index().id();
-    if (element_size_log2 > 0) os << "*" << (1 << element_size_log2);
+    os << " + " << op_index_prefix << index().value().id();
+    if (element_size_log2 > 0) os << '*' << (1 << element_size_log2);
   }
   os << ") ";
 }
 void LoadOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   os << (kind.tagged_base ? "tagged base" : "raw");
   if (kind.maybe_unaligned) os << ", unaligned";
   if (kind.with_trap_handler) os << ", protected";
   os << ", " << loaded_rep;
+  os << ", " << result_rep;
   if (element_size_log2 != 0)
     os << ", element size: 2^" << int{element_size_log2};
   if (offset != 0) os << ", offset: " << offset;
-  os << "]";
+  os << ']';
 }
 
 void AtomicRMWOp::PrintInputs(std::ostream& os,
                               const std::string& op_index_prefix) const {
   os << " *(" << op_index_prefix << base().id() << " + " << op_index_prefix
-     << index().id() << ").atomic_" << bin_op << "(";
+     << index().id() << ").atomic_" << bin_op << '(';
   if (bin_op == BinOp::kCompareExchange) {
-    os << "expected: " << op_index_prefix << expected().id();
-    os << ", new: " << op_index_prefix << value().id();
+    os << "expected: " << op_index_prefix << expected();
+    os << ", new: " << op_index_prefix << value();
   } else {
     os << op_index_prefix << value().id();
   }
-  os << ")";
+  os << ')';
 }
 
 void AtomicRMWOp::PrintOptions(std::ostream& os) const {
-  os << "["
-     << "binop: " << bin_op << ", result_rep: " << result_rep
-     << ", input_rep: " << input_rep << "]";
+  os << '[' << "binop: " << bin_op << ", in_out_rep: " << in_out_rep
+     << ", memory_rep: " << memory_rep << ']';
 }
 
 void AtomicWord32PairOp::PrintInputs(std::ostream& os,
                                      const std::string& op_index_prefix) const {
   os << " *(" << op_index_prefix << base().id();
   if (index().valid()) {
-    os << " + " << op_index_prefix << index().id();
+    os << " + " << op_index_prefix << index().value().id();
   }
   if (offset) {
     os << " + offset=" << offset;
   }
-  os << ").atomic_word32_pair_" << op_kind << "(";
-  if (op_kind == OpKind::kCompareExchange) {
-    os << "expected: {lo: " << op_index_prefix << value_low().id()
+  os << ").atomic_word32_pair_" << kind << '(';
+  if (kind == Kind::kCompareExchange) {
+    os << "expected: {lo: " << op_index_prefix << value_low()
        << ", hi: " << op_index_prefix << value_high();
-    os << "}, value: {lo: " << op_index_prefix << value_low().id()
-       << ", hi: " << op_index_prefix << value_high() << "}";
-  } else if (op_kind != OpKind::kLoad) {
-    os << "lo: " << op_index_prefix << value_low().id()
+    os << "}, value: {lo: " << op_index_prefix << value_low()
+       << ", hi: " << op_index_prefix << value_high() << '}';
+  } else if (kind != Kind::kLoad) {
+    os << "lo: " << op_index_prefix << value_low()
        << ", hi: " << op_index_prefix << value_high();
   }
-  os << ")";
+  os << ')';
 }
 
 void AtomicWord32PairOp::PrintOptions(std::ostream& os) const {
-  os << "[opkind: " << op_kind << "]";
+  os << "[opkind: " << kind << ']';
+}
+
+void MemoryBarrierOp::PrintOptions(std::ostream& os) const {
+  os << "[memory order: " << memory_order << ']';
 }
 
 void StoreOp::PrintInputs(std::ostream& os,
@@ -540,13 +678,13 @@ void StoreOp::PrintInputs(std::ostream& os,
     os << " + " << offset;
   }
   if (index().valid()) {
-    os << " + " << op_index_prefix << index().id();
-    if (element_size_log2 > 0) os << "*" << (1 << element_size_log2);
+    os << " + " << op_index_prefix << index().value().id();
+    if (element_size_log2 > 0) os << '*' << (1 << element_size_log2);
   }
-  os << ") = " << op_index_prefix << value().id() << " ";
+  os << ") = " << op_index_prefix << value().id() << ' ';
 }
 void StoreOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   os << (kind.tagged_base ? "tagged base" : "raw");
   if (kind.maybe_unaligned) os << ", unaligned";
   if (kind.with_trap_handler) os << ", protected";
@@ -556,59 +694,59 @@ void StoreOp::PrintOptions(std::ostream& os) const {
     os << ", element size: 2^" << int{element_size_log2};
   if (offset != 0) os << ", offset: " << offset;
   if (maybe_initializing_or_transitioning) os << ", initializing";
-  os << "]";
+  os << ']';
 }
 
 void AllocateOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   os << type;
-  os << "]";
+  os << ']';
 }
 
 void DecodeExternalPointerOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   os << "tag: " << std::hex << tag << std::dec;
-  os << "]";
+  os << ']';
 }
 
 void FrameStateOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   os << (inlined ? "inlined" : "not inlined");
   os << ", ";
   os << data->frame_state_info;
   os << ", state values:";
   FrameStateData::Iterator it = data->iterator(state_values());
   while (it.has_more()) {
-    os << " ";
+    os << ' ';
     switch (it.current_instr()) {
       case FrameStateData::Instr::kInput: {
         MachineType type;
         OpIndex input;
         it.ConsumeInput(&type, &input);
-        os << "#" << input.id() << "(" << type << ")";
+        os << '#' << input.id() << '(' << type << ')';
         break;
       }
       case FrameStateData::Instr::kUnusedRegister:
         it.ConsumeUnusedRegister();
-        os << ".";
+        os << '.';
         break;
       case FrameStateData::Instr::kDematerializedObject: {
         uint32_t id;
         uint32_t field_count;
         it.ConsumeDematerializedObject(&id, &field_count);
-        os << "$" << id << "(field count: " << field_count << ")";
+        os << '$' << id << "(field count: " << field_count << ')';
         break;
       }
       case FrameStateData::Instr::kDematerializedObjectReference: {
         uint32_t id;
         it.ConsumeDematerializedObjectReference(&id);
-        os << "$" << id;
+        os << '$' << id;
         break;
       }
       case FrameStateData::Instr::kArgumentsElements: {
         CreateArgumentsType type;
         it.ConsumeArgumentsElements(&type);
-        os << "ArgumentsElements(" << type << ")";
+        os << "ArgumentsElements(" << type << ')';
         break;
       }
       case FrameStateData::Instr::kArgumentsLength: {
@@ -616,9 +754,14 @@ void FrameStateOp::PrintOptions(std::ostream& os) const {
         os << "ArgumentsLength";
         break;
       }
+      case FrameStateData::Instr::kRestLength: {
+        it.ConsumeRestLength();
+        os << "RestLength";
+        break;
+      }
     }
   }
-  os << "]";
+  os << ']';
 }
 
 void FrameStateOp::Validate(const Graph& graph) const {
@@ -665,19 +808,36 @@ void FrameStateOp::Validate(const Graph& graph) const {
         it.ConsumeArgumentsLength();
         break;
       }
+      case FrameStateData::Instr::kRestLength: {
+        it.ConsumeRestLength();
+        break;
+      }
     }
   }
 }
 
+void DeoptimizeIfOp::PrintOptions(std::ostream& os) const {
+  static_assert(std::tuple_size_v<decltype(options())> == 2);
+  os << '[' << (negated ? "negated, " : "") << *parameters << ']';
+}
+
 void DidntThrowOp::Validate(const Graph& graph) const {
 #ifdef DEBUG
   DCHECK(MayThrow(graph.Get(throwing_operation()).opcode));
   switch (graph.Get(throwing_operation()).opcode) {
     case Opcode::kCall: {
       auto& call_op = graph.Get(throwing_operation()).Cast<CallOp>();
-      DCHECK(call_op.descriptor->out_reps == outputs_rep());
+      DCHECK_EQ(call_op.descriptor->out_reps, outputs_rep());
       break;
     }
+#define STATIC_OUTPUT_CASE(Name)                                           \
+  case Opcode::k##Name: {                                                  \
+    const Name##Op& op = graph.Get(throwing_operation()).Cast<Name##Op>(); \
+    DCHECK_EQ(op.kOutReps, outputs_rep());                                 \
+    break;                                                                 \
+  }
+      TURBOSHAFT_THROWING_STATIC_OUTPUTS_OPERATIONS_LIST(STATIC_OUTPUT_CASE)
+#undef STATIC_OUTPUT_CASE
     default:
       UNREACHABLE();
   }
@@ -696,7 +856,7 @@ void DidntThrowOp::Validate(const Graph& graph) const {
 }
 
 void WordBinopOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kAdd:
       os << "Add, ";
@@ -736,11 +896,11 @@ void WordBinopOp::PrintOptions(std::ostream& os) const {
       break;
   }
   os << rep;
-  os << "]";
+  os << ']';
 }
 
 void FloatBinopOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kAdd:
       os << "Add, ";
@@ -771,11 +931,11 @@ void FloatBinopOp::PrintOptions(std::ostream& os) const {
       break;
   }
   os << rep;
-  os << "]";
+  os << ']';
 }
 
 void Word32PairBinopOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kAdd:
       os << "Add";
@@ -796,11 +956,40 @@ void Word32PairBinopOp::PrintOptions(std::ostream& os) const {
       os << "ShiftRightUnsigned";
       break;
   }
-  os << "]";
+  os << ']';
+}
+
+void WordBinopDeoptOnOverflowOp::PrintOptions(std::ostream& os) const {
+  os << '[';
+  switch (kind) {
+    case Kind::kSignedAdd:
+      os << "signed add, ";
+      break;
+    case Kind::kSignedMul:
+      os << "signed mul, ";
+      break;
+    case Kind::kSignedSub:
+      os << "signed sub, ";
+      break;
+    case Kind::kSignedDiv:
+      os << "signed div, ";
+      break;
+    case Kind::kSignedMod:
+      os << "signed mod, ";
+      break;
+    case Kind::kUnsignedDiv:
+      os << "unsigned div, ";
+      break;
+    case Kind::kUnsignedMod:
+      os << "unsigned mod, ";
+      break;
+  }
+  os << rep << ", " << mode;
+  os << ']';
 }
 
 void OverflowCheckedBinopOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kSignedAdd:
       os << "signed add, ";
@@ -813,7 +1002,7 @@ void OverflowCheckedBinopOp::PrintOptions(std::ostream& os) const {
       break;
   }
   os << rep;
-  os << "]";
+  os << ']';
 }
 
 std::ostream& operator<<(std::ostream& os, OpIndex idx) {
@@ -865,17 +1054,17 @@ std::ostream& operator<<(std::ostream& os, OpEffects effects) {
   os << produce_consume(effects.produces.control_flow,
                         effects.consumes.control_flow);
   os << "\u2003";  // em space
-  os << (effects.can_create_identity ? "i" : "_");
-  os << " " << (effects.can_allocate ? "a" : "_");
+  os << (effects.can_create_identity ? 'i' : '_');
+  os << ' ' << (effects.can_allocate ? 'a' : '_');
   return os;
 }
 
 void SwitchOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   for (const Case& c : cases) {
     os << "case " << c.value << ": " << c.destination << ", ";
   }
-  os << " default: " << default_case << "]";
+  os << " default: " << default_case << ']';
 }
 
 std::ostream& operator<<(std::ostream& os, ObjectIsOp::Kind kind) {
@@ -898,6 +1087,8 @@ std::ostream& operator<<(std::ostream& os, ObjectIsOp::Kind kind) {
       return os << "NonCallable";
     case ObjectIsOp::Kind::kNumber:
       return os << "Number";
+    case ObjectIsOp::Kind::kNumberOrBigInt:
+      return os << "NumberOrBigInt";
     case ObjectIsOp::Kind::kReceiver:
       return os << "Receiver";
     case ObjectIsOp::Kind::kReceiverOrNullOrUndefined:
@@ -906,6 +1097,8 @@ std::ostream& operator<<(std::ostream& os, ObjectIsOp::Kind kind) {
       return os << "Smi";
     case ObjectIsOp::Kind::kString:
       return os << "String";
+    case ObjectIsOp::Kind::kStringOrStringWrapper:
+      return os << "StringOrStringWrapper";
     case ObjectIsOp::Kind::kSymbol:
       return os << "Symbol";
     case ObjectIsOp::Kind::kUndetectable:
@@ -935,6 +1128,8 @@ std::ostream& operator<<(std::ostream& os, NumericKind kind) {
       return os << "Integer";
     case NumericKind::kSafeInteger:
       return os << "SafeInteger";
+    case NumericKind::kSmi:
+      return os << "kSmi";
     case NumericKind::kMinusZero:
       return os << "MinusZero";
     case NumericKind::kNaN:
@@ -970,6 +1165,9 @@ std::ostream& operator<<(std::ostream& os,
       return os << "Boolean";
     case ConvertUntaggedToJSPrimitiveOp::JSPrimitiveKind::kHeapNumber:
       return os << "HeapNumber";
+    case ConvertUntaggedToJSPrimitiveOp::JSPrimitiveKind::
+        kHeapNumberOrUndefined:
+      return os << "HeapNumberOrUndefined";
     case ConvertUntaggedToJSPrimitiveOp::JSPrimitiveKind::kNumber:
       return os << "Number";
     case ConvertUntaggedToJSPrimitiveOp::JSPrimitiveKind::kSmi:
@@ -1161,6 +1359,8 @@ std::ostream& operator<<(std::ostream& os, BigIntBinopOp::Kind kind) {
 
 std::ostream& operator<<(std::ostream& os, BigIntComparisonOp::Kind kind) {
   switch (kind) {
+    case BigIntComparisonOp::Kind::kEqual:
+      return os << "Equal";
     case BigIntComparisonOp::Kind::kLessThan:
       return os << "LessThan";
     case BigIntComparisonOp::Kind::kLessThanOrEqual:
@@ -1197,6 +1397,8 @@ std::ostream& operator<<(std::ostream& os, StringToCaseIntlOp::Kind kind) {
 
 std::ostream& operator<<(std::ostream& os, StringComparisonOp::Kind kind) {
   switch (kind) {
+    case StringComparisonOp::Kind::kEqual:
+      return os << "Equal";
     case StringComparisonOp::Kind::kLessThan:
       return os << "LessThan";
     case StringComparisonOp::Kind::kLessThanOrEqual:
@@ -1249,6 +1451,25 @@ std::ostream& operator<<(std::ostream& os, FindOrderedHashEntryOp::Kind kind) {
   }
 }
 
+std::ostream& operator<<(std::ostream& os,
+                         SpeculativeNumberBinopOp::Kind kind) {
+  switch (kind) {
+    case SpeculativeNumberBinopOp::Kind::kSafeIntegerAdd:
+      return os << "SafeIntegerAdd";
+  }
+}
+
+std::ostream& operator<<(std::ostream& os, JSStackCheckOp::Kind kind) {
+  switch (kind) {
+    case JSStackCheckOp::Kind::kFunctionEntry:
+      return os << "function-entry";
+    case JSStackCheckOp::Kind::kBuiltinEntry:
+      return os << "builtin-entry";
+    case JSStackCheckOp::Kind::kLoop:
+      return os << "loop";
+  }
+}
+
 #if V8_ENABLE_WEBASSEMBLY
 
 const RegisterRepresentation& RepresentationFor(wasm::ValueType type) {
@@ -1272,6 +1493,7 @@ const RegisterRepresentation& RepresentationFor(wasm::ValueType type) {
       return kWord32;
     case wasm::kI64:
       return kWord64;
+    case wasm::kF16:
     case wasm::kF32:
       return kFloat32;
     case wasm::kF64:
@@ -1289,12 +1511,13 @@ const RegisterRepresentation& RepresentationFor(wasm::ValueType type) {
 }
 
 namespace {
-void PrintSimd128Value(std::ostream& os, const uint8_t value[kSimd128Size]) {
+template <size_t size>
+void PrintSimdValue(std::ostream& os, const uint8_t (&value)[size]) {
   os << "0x" << std::hex << std::setfill('0');
 #ifdef V8_TARGET_BIG_ENDIAN
-  for (int i = 0; i < kSimd128Size; i++) {
+  for (int i = 0; i < static_cast<int>(size); i++) {
 #else
-  for (int i = kSimd128Size - 1; i >= 0; i--) {
+  for (int i = static_cast<int>(size) - 1; i >= 0; i--) {
 #endif
     os << std::setw(2) << static_cast<int>(value[i]);
   }
@@ -1303,7 +1526,7 @@ void PrintSimd128Value(std::ostream& os, const uint8_t value[kSimd128Size]) {
 }  // namespace
 
 void Simd128ConstantOp::PrintOptions(std::ostream& os) const {
-  PrintSimd128Value(os, value);
+  PrintSimdValue(os, value);
 }
 
 std::ostream& operator<<(std::ostream& os, Simd128BinopOp::Kind kind) {
@@ -1326,6 +1549,16 @@ std::ostream& operator<<(std::ostream& os, Simd128UnaryOp::Kind kind) {
 #undef PRINT_KIND
 }
 
+std::ostream& operator<<(std::ostream& os, Simd128ReduceOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)               \
+  case Simd128ReduceOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_128_REDUCE_OPTIONAL_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+
 std::ostream& operator<<(std::ostream& os, Simd128ShiftOp::Kind kind) {
   switch (kind) {
 #define PRINT_KIND(kind)              \
@@ -1367,7 +1600,7 @@ std::ostream& operator<<(std::ostream& os, Simd128TernaryOp::Kind kind) {
 }
 
 void Simd128ExtractLaneOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kI8x16S:
       os << "I8x16S";
@@ -1387,6 +1620,9 @@ void Simd128ExtractLaneOp::PrintOptions(std::ostream& os) const {
     case Kind::kI64x2:
       os << "I64x2";
       break;
+    case Kind::kF16x8:
+      os << "F16x8";
+      break;
     case Kind::kF32x4:
       os << "F32x4";
       break;
@@ -1394,11 +1630,11 @@ void Simd128ExtractLaneOp::PrintOptions(std::ostream& os) const {
       os << "F64x2";
       break;
   }
-  os << ", " << static_cast<int32_t>(lane) << "]";
+  os << ", " << static_cast<int32_t>(lane) << ']';
 }
 
 void Simd128ReplaceLaneOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   switch (kind) {
     case Kind::kI8x16:
       os << "I8x16";
@@ -1412,6 +1648,9 @@ void Simd128ReplaceLaneOp::PrintOptions(std::ostream& os) const {
     case Kind::kI64x2:
       os << "I64x2";
       break;
+    case Kind::kF16x8:
+      os << "F16x8";
+      break;
     case Kind::kF32x4:
       os << "F32x4";
       break;
@@ -1419,16 +1658,16 @@ void Simd128ReplaceLaneOp::PrintOptions(std::ostream& os) const {
       os << "F64x2";
       break;
   }
-  os << ", " << static_cast<int32_t>(lane) << "]";
+  os << ", " << static_cast<int32_t>(lane) << ']';
 }
 
 void Simd128LaneMemoryOp::PrintOptions(std::ostream& os) const {
-  os << "[" << (mode == Mode::kLoad ? "Load" : "Store") << ", ";
+  os << '[' << (mode == Mode::kLoad ? "Load" : "Store") << ", ";
   if (kind.maybe_unaligned) os << "unaligned, ";
   if (kind.with_trap_handler) os << "protected, ";
   switch (lane_kind) {
     case LaneKind::k8:
-      os << "8";
+      os << '8';
       break;
     case LaneKind::k16:
       os << "16";
@@ -1442,11 +1681,11 @@ void Simd128LaneMemoryOp::PrintOptions(std::ostream& os) const {
   }
   os << "bit, lane: " << static_cast<int>(lane);
   if (offset != 0) os << ", offset: " << offset;
-  os << "]";
+  os << ']';
 }
 
 void Simd128LoadTransformOp::PrintOptions(std::ostream& os) const {
-  os << "[";
+  os << '[';
   if (load_kind.maybe_unaligned) os << "unaligned, ";
   if (load_kind.with_trap_handler) os << "protected, ";
 
@@ -1459,11 +1698,118 @@ void Simd128LoadTransformOp::PrintOptions(std::ostream& os) const {
 #undef PRINT_KIND
   }
 
-  os << ", offset: " << offset << "]";
+  os << ", offset: " << offset << ']';
 }
 
 void Simd128ShuffleOp::PrintOptions(std::ostream& os) const {
-  PrintSimd128Value(os, shuffle);
+  PrintSimdValue(os, shuffle);
+}
+
+#if V8_ENABLE_WASM_SIMD256_REVEC
+void Simd256ConstantOp::PrintOptions(std::ostream& os) const {
+  PrintSimdValue(os, value);
+}
+
+void Simd256Extract128LaneOp::PrintOptions(std::ostream& os) const {
+  os << '[' << static_cast<int>(lane) << ']';
+}
+
+void Simd256LoadTransformOp::PrintOptions(std::ostream& os) const {
+  os << '[';
+  if (load_kind.maybe_unaligned) os << "unaligned, ";
+  if (load_kind.with_trap_handler) os << "protected, ";
+
+  switch (transform_kind) {
+#define PRINT_KIND(kind)       \
+  case TransformKind::k##kind: \
+    os << #kind;               \
+    break;
+    FOREACH_SIMD_256_LOAD_TRANSFORM_OPCODE(PRINT_KIND)
+#undef PRINT_KIND
+  }
+
+  os << ", offset: " << offset << ']';
+}
+
+std::ostream& operator<<(std::ostream& os, Simd256UnaryOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)              \
+  case Simd256UnaryOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_256_UNARY_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+
+std::ostream& operator<<(std::ostream& os, Simd256TernaryOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)                \
+  case Simd256TernaryOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_256_TERNARY_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+
+std::ostream& operator<<(std::ostream& os, Simd256BinopOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)              \
+  case Simd256BinopOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_256_BINARY_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+
+std::ostream& operator<<(std::ostream& os, Simd256ShiftOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)              \
+  case Simd256ShiftOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_256_SHIFT_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+
+std::ostream& operator<<(std::ostream& os, Simd256SplatOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)              \
+  case Simd256SplatOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_256_SPLAT_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+
+#ifdef V8_TARGET_ARCH_X64
+void Simd256ShufdOp::PrintOptions(std::ostream& os) const {
+  os << '[' << std::bitset<8>(control) << ']';
+}
+
+void Simd256ShufpsOp::PrintOptions(std::ostream& os) const {
+  os << '[' << std::bitset<8>(control) << ']';
+}
+
+std::ostream& operator<<(std::ostream& os, Simd256UnpackOp::Kind kind) {
+  switch (kind) {
+#define PRINT_KIND(kind)               \
+  case Simd256UnpackOp::Kind::k##kind: \
+    return os << #kind;
+    FOREACH_SIMD_256_UNPACK_OPCODE(PRINT_KIND)
+  }
+#undef PRINT_KIND
+}
+#endif  // V8_TARGET_ARCH_X64
+#endif  // V8_ENABLE_WASM_SIMD256_REVEC
+
+void WasmAllocateArrayOp::PrintOptions(std::ostream& os) const {
+  os << '[' << array_type->element_type() << ']';
+}
+
+void ArrayGetOp::PrintOptions(std::ostream& os) const {
+  os << '[' << (is_signed ? "signed " : "")
+     << (array_type->mutability() ? "" : "immutable ")
+     << array_type->element_type() << ']';
 }
 
 #endif  // V8_ENABLE_WEBASSEBMLY
@@ -1512,7 +1858,82 @@ bool SupportedOperations::IsUnalignedStoreSupported(MemoryRepresentation repr) {
 void CheckExceptionOp::Validate(const Graph& graph) const {
   DCHECK_NE(didnt_throw_block, catch_block);
   // `CheckException` should follow right after the throwing operation.
-  DCHECK_EQ(throwing_operation(), graph.PreviousIndex(graph.Index(*this)));
+  DCHECK_EQ(throwing_operation(),
+            V<Any>::Cast(graph.PreviousIndex(graph.Index(*this))));
+}
+
+namespace {
+BlockIndex index_for_bound_block(const Block* block) {
+  DCHECK_NOT_NULL(block);
+  const BlockIndex index = block->index();
+  DCHECK(index.valid());
+  return index;
+}
+}  // namespace
+
+size_t CallOp::hash_value(HashingStrategy strategy) const {
+  if (strategy == HashingStrategy::kMakeSnapshotStable) {
+    // Destructure here to cause a compilation error in case `options` is
+    // changed.
+    auto [descriptor_value, callee_effects] = options();
+    return HashWithOptions(*descriptor_value, callee_effects);
+  } else {
+    return Base::hash_value(strategy);
+  }
+}
+
+size_t CheckExceptionOp::hash_value(HashingStrategy strategy) const {
+  if (strategy == HashingStrategy::kMakeSnapshotStable) {
+    // Destructure here to cause a compilation error in case `options` is
+    // changed.
+    auto [didnt_throw_block_value, catch_block_value] = options();
+    return HashWithOptions(index_for_bound_block(didnt_throw_block_value),
+                           index_for_bound_block(catch_block_value));
+  } else {
+    return Base::hash_value(strategy);
+  }
+}
+
+size_t GotoOp::hash_value(HashingStrategy strategy) const {
+  if (strategy == HashingStrategy::kMakeSnapshotStable) {
+    // Destructure here to cause a compilation error in case `options` is
+    // changed.
+    auto [destination_value, is_backedge_value] = options();
+    return HashWithOptions(index_for_bound_block(destination_value),
+                           is_backedge_value);
+  } else {
+    return Base::hash_value(strategy);
+  }
+}
+
+size_t BranchOp::hash_value(HashingStrategy strategy) const {
+  if (strategy == HashingStrategy::kMakeSnapshotStable) {
+    // Destructure here to cause a compilation error in case `options` is
+    // changed.
+    auto [if_true_value, if_false_value, hint_value] = options();
+    return HashWithOptions(index_for_bound_block(if_true_value),
+                           index_for_bound_block(if_false_value), hint_value);
+  } else {
+    return Base::hash_value(strategy);
+  }
+}
+
+size_t SwitchOp::hash_value(HashingStrategy strategy) const {
+  if (strategy == HashingStrategy::kMakeSnapshotStable) {
+    // Destructure here to cause a compilation error in case `options` is
+    // changed.
+    auto [cases_value, default_case_value, default_hint_value] = options();
+    DCHECK_NOT_NULL(default_case_value);
+    size_t hash = HashWithOptions(index_for_bound_block(default_case_value),
+                                  default_hint_value);
+    for (const auto& c : cases_value) {
+      hash = fast_hash_combine(hash, c.value,
+                               index_for_bound_block(c.destination), c.hint);
+    }
+    return hash;
+  } else {
+    return Base::hash_value(strategy);
+  }
 }
 
 namespace {
@@ -1578,4 +1999,12 @@ bool IsUnlikelySuccessor(const Block* block, const Block* successor,
   }
 }
 
+bool Operation::IsOnlyUserOf(const Operation& value, const Graph& graph) const {
+  DCHECK_GE(std::count(inputs().begin(), inputs().end(), graph.Index(value)),
+            1);
+  if (value.saturated_use_count.IsOne()) return true;
+  return std::count(inputs().begin(), inputs().end(), graph.Index(value)) ==
+         value.saturated_use_count.Get();
+}
+
 }  // namespace v8::internal::compiler::turboshaft
