@@ -131,7 +131,7 @@ struct ParserTypes<Parser> {
 
 class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
  public:
-  Parser(LocalIsolate* local_isolate, ParseInfo* info, Handle<Script> script);
+  Parser(LocalIsolate* local_isolate, ParseInfo* info);
   ~Parser() {
     delete reusable_preparser_;
     reusable_preparser_ = nullptr;
@@ -141,8 +141,8 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
 
   // Sets the literal on |info| if parsing succeeded.
   void ParseOnBackground(LocalIsolate* isolate, ParseInfo* info,
-                         int start_position, int end_position,
-                         int function_literal_id);
+                         DirectHandle<Script> script, int start_position,
+                         int end_position, int function_literal_id);
 
   // Initializes an empty scope chain for top-level scripts, or scopes which
   // consist of only the native context.
@@ -163,13 +163,13 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                                  Scope::DeserializationMode::kScopesOnly);
 
   // Move statistics to Isolate
-  void UpdateStatistics(Isolate* isolate, Handle<Script> script);
+  void UpdateStatistics(Isolate* isolate, DirectHandle<Script> script);
   void UpdateStatistics(
-      Handle<Script> script,
+      DirectHandle<Script> script,
       base::SmallVector<v8::Isolate::UseCounterFeature, 8>* use_counters,
       int* preparse_skipped);
   template <typename IsolateT>
-  void HandleSourceURLComments(IsolateT* isolate, Handle<Script> script);
+  void HandleSourceURLComments(IsolateT* isolate, DirectHandle<Script> script);
 
  private:
   friend class ParserBase<Parser>;
@@ -179,8 +179,9 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
   friend class i::ParameterDeclarationParsingScope<ParserTypes<Parser>>;
   friend class i::ArrowHeadParsingScope<ParserTypes<Parser>>;
   friend bool v8::internal::parsing::ParseProgram(
-      ParseInfo*, Handle<Script>, MaybeHandle<ScopeInfo> maybe_outer_scope_info,
-      Isolate*, parsing::ReportStatisticsMode stats_mode);
+      ParseInfo*, DirectHandle<Script>,
+      MaybeHandle<ScopeInfo> maybe_outer_scope_info, Isolate*,
+      parsing::ReportStatisticsMode stats_mode);
   friend bool v8::internal::parsing::ParseFunction(
       ParseInfo*, Handle<SharedFunctionInfo> shared_info, Isolate*,
       parsing::ReportStatisticsMode stats_mode);
@@ -221,12 +222,13 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
   void PrepareGeneratorVariables();
 
   // Sets the literal on |info| if parsing succeeded.
-  void ParseProgram(Isolate* isolate, Handle<Script> script, ParseInfo* info,
+  void ParseProgram(Isolate* isolate, DirectHandle<Script> script,
+                    ParseInfo* info,
                     MaybeHandle<ScopeInfo> maybe_outer_scope_info);
 
   // Sets the literal on |info| if parsing succeeded.
   void ParseFunction(Isolate* isolate, ParseInfo* info,
-                     Handle<SharedFunctionInfo> shared_info);
+                     DirectHandle<SharedFunctionInfo> shared_info);
 
   template <typename IsolateT>
   void PostProcessParseResult(IsolateT* isolate, ParseInfo* info,
@@ -238,7 +240,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                                    const AstRawString* raw_name);
 
   FunctionLiteral* ParseClassForMemberInitialization(
-      Isolate* isolate, MaybeHandle<ScopeInfo> maybe_class_scope_info,
       FunctionKind initalizer_kind, int initializer_pos, int initializer_id,
       int initializer_end_pos, const AstRawString* class_name);
 
@@ -300,9 +301,8 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
   const AstRawString* ParseExportSpecifierName();
   ZonePtrList<const NamedImport>* ParseNamedImports(int pos);
 
-  ImportAssertions* ParseImportAssertClause();
+  ImportAttributes* ParseImportWithOrAssertClause();
   Statement* BuildInitializationBlock(DeclarationParsingResult* parsing_result);
-  Expression* RewriteReturn(Expression* return_value, int pos);
   Statement* RewriteSwitchStatement(SwitchStatement* switch_statement,
                                     Scope* scope);
   Block* RewriteCatchPattern(CatchInfo* catch_info);
@@ -312,8 +312,8 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                                  Block* finally_block,
                                  const SourceRange& finally_range,
                                  const CatchInfo& catch_info, int pos);
-  void ParseAndRewriteGeneratorFunctionBody(int pos, FunctionKind kind,
-                                            ScopedPtrList<Statement>* body);
+  void ParseGeneratorFunctionBody(int pos, FunctionKind kind,
+                                  ScopedPtrList<Statement>* body);
   void ParseAndRewriteAsyncGeneratorFunctionBody(
       int pos, FunctionKind kind, ScopedPtrList<Statement>* body);
   void DeclareFunctionNameVar(const AstRawString* function_name,
@@ -324,12 +324,16 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                              FunctionLiteral* function, VariableMode mode,
                              VariableKind kind, int beg_pos, int end_pos,
                              ZonePtrList<const AstRawString>* names);
-  Variable* CreateSyntheticContextVariable(const AstRawString* synthetic_name);
-  Variable* CreatePrivateNameVariable(ClassScope* scope, VariableMode mode,
-                                      IsStaticFlag is_static_flag,
-                                      const AstRawString* name);
+  VariableProxy* CreateSyntheticContextVariableProxy(ClassScope* scope,
+                                                     ClassInfo* class_info,
+                                                     const AstRawString* name,
+                                                     bool is_static);
+  VariableProxy* CreatePrivateNameVariable(ClassScope* scope, VariableMode mode,
+                                           IsStaticFlag is_static_flag,
+                                           const AstRawString* name);
   FunctionLiteral* CreateInitializerFunction(const AstRawString* class_name,
                                              DeclarationScope* scope,
+                                             int function_literal_id,
                                              Statement* initializer_stmt);
 
   bool IdentifierEquals(const AstRawString* identifier,
@@ -363,9 +367,13 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                          bool is_computed_name, bool is_private,
                          ClassInfo* class_info);
   void AddClassStaticBlock(Block* block, ClassInfo* class_info);
+  FunctionLiteral* CreateStaticElementsInitializer(const AstRawString* name,
+                                                   ClassInfo* class_info);
+  FunctionLiteral* CreateInstanceMembersInitializer(const AstRawString* name,
+                                                    ClassInfo* class_info);
   Expression* RewriteClassLiteral(ClassScope* block_scope,
                                   const AstRawString* name,
-                                  ClassInfo* class_info, int pos, int end_pos);
+                                  ClassInfo* class_info, int pos);
   Statement* DeclareNative(const AstRawString* name, int pos);
 
   Block* IgnoreCompletion(Statement* statement);
@@ -428,7 +436,24 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
 
   // Factory methods.
   FunctionLiteral* DefaultConstructor(const AstRawString* name, bool call_super,
-                                      int pos, int end_pos);
+                                      int pos);
+
+  FunctionLiteral* MakeAutoAccessorGetter(VariableProxy* name_proxy,
+                                          const AstRawString* name,
+                                          bool is_static, int pos);
+
+  FunctionLiteral* MakeAutoAccessorSetter(VariableProxy* name_proxy,
+                                          const AstRawString* name,
+                                          bool is_static, int pos);
+
+  AutoAccessorInfo* NewAutoAccessorInfo(ClassScope* scope,
+                                        ClassInfo* class_info,
+                                        const AstRawString* name,
+                                        bool is_static, int pos);
+  ClassLiteralProperty* NewClassLiteralPropertyWithAccessorInfo(
+      ClassScope* scope, ClassInfo* class_info, const AstRawString* name,
+      Expression* key, Expression* value, bool is_static, bool is_computed_name,
+      bool is_private, int pos);
 
   // Skip over a lazy function, either using cached data if we have it, or
   // by parsing the function with PreParser. Consumes the ending }.
@@ -447,8 +472,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
 
   Block* BuildParameterInitializationBlock(
       const ParserFormalParameters& parameters);
-  Block* BuildRejectPromiseOnException(Block* block,
-                                       REPLMode repl_mode = REPLMode::kNo);
 
   void ParseFunction(
       ScopedPtrList<Statement>* body, const AstRawString* function_name,
@@ -504,8 +527,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
   Expression* CloseTemplateLiteral(TemplateLiteralState* state, int start,
                                    Expression* tag);
 
-  ArrayLiteral* ArrayLiteralFromListWithSpread(
-      const ScopedPtrList<Expression>& list);
   Expression* RewriteSuperCall(Expression* call_expression);
 
   void SetLanguageMode(Scope* scope, LanguageMode mode);
@@ -523,12 +544,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                             MessageTemplate message, const AstRawString* arg,
                             int pos);
 
-  Statement* CheckCallable(Variable* var, Expression* error, int pos);
-
-  void RewriteAsyncFunctionBody(ScopedPtrList<Statement>* body, Block* block,
-                                Expression* return_value,
-                                REPLMode repl_mode = REPLMode::kNo);
-
   void AddArrowFunctionFormalParameters(ParserFormalParameters* parameters,
                                         Expression* params, int end_pos);
   void SetFunctionName(Expression* value, const AstRawString* name,
@@ -590,10 +605,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
     return identifier == ast_value_factory()->constructor_string();
   }
 
-  V8_INLINE bool IsName(const AstRawString* identifier) const {
-    return identifier == ast_value_factory()->name_string();
-  }
-
   V8_INLINE static bool IsBoilerplateProperty(
       ObjectLiteral::Property* property) {
     return !property->IsPrototype();
@@ -671,11 +682,18 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
     }
   }
 
-  // Returns true if we have a binary expression between two numeric
-  // literals. In that case, *x will be changed to an expression which is the
-  // computed value.
-  bool ShortcutNumericLiteralBinaryExpression(Expression** x, Expression* y,
-                                              Token::Value op, int pos);
+  // Returns true if we have a binary expression between two literals. In that
+  // case, *x will be changed to an expression which is the computed value.
+  bool ShortcutLiteralBinaryExpression(Expression** x, Expression* y,
+                                       Token::Value op, int pos);
+
+  bool CollapseConditionalChain(Expression** x, Expression* cond,
+                                Expression* then_expression,
+                                Expression* else_expression, int pos,
+                                const SourceRange& then_range);
+
+  void AppendConditionalChainElse(Expression** x,
+                                  const SourceRange& else_range);
 
   // Returns true if we have a binary operation between a binary/n-ary
   // expression (with the same operation) and a value, which can be collapsed
@@ -699,21 +717,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                          ast_value_factory()->empty_string(), pos);
   }
 
-  // Generate AST node that throws a SyntaxError with the given
-  // type. The first argument may be null (in the handle sense) in
-  // which case no arguments are passed to the constructor.
-  V8_INLINE Expression* NewThrowSyntaxError(MessageTemplate message,
-                                            const AstRawString* arg, int pos) {
-    return NewThrowError(Runtime::kNewSyntaxError, message, arg, pos);
-  }
-
-  // Generate AST node that throws a TypeError with the given
-  // type. Both arguments must be non-null (in the handle sense).
-  V8_INLINE Expression* NewThrowTypeError(MessageTemplate message,
-                                          const AstRawString* arg, int pos) {
-    return NewThrowError(Runtime::kNewTypeError, message, arg, pos);
-  }
-
   // Dummy implementation. The parser should never have a unidentifiable
   // error.
   V8_INLINE void ReportUnidentifiableError() { UNREACHABLE(); }
@@ -764,6 +767,10 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
   V8_INLINE const AstRawString* EmptyIdentifierString() const {
     return ast_value_factory()->empty_string();
   }
+  V8_INLINE bool IsEmptyIdentifier(const AstRawString* subject) const {
+    DCHECK_NOT_NULL(subject);
+    return subject->IsEmpty();
+  }
 
   // Producing data during the recursive descent.
   V8_INLINE const AstRawString* GetSymbol() const {
@@ -798,8 +805,8 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
     return factory()->NewThisExpression(pos);
   }
 
-  Expression* NewSuperPropertyReference(Scope* home_object_scope, int pos);
-  Expression* NewSuperCallReference(int pos);
+  Expression* NewSuperPropertyReference(int pos);
+  SuperCallReference* NewSuperCallReference(int pos);
   Expression* NewTargetExpression(int pos);
   Expression* ImportMetaExpression(int pos);
 
@@ -833,13 +840,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
     return scope->DeclareCatchVariableName(name);
   }
 
-  V8_INLINE ZonePtrList<Expression>* NewExpressionList(int size) const {
-    return zone()->New<ZonePtrList<Expression>>(size, zone());
-  }
-  V8_INLINE ZonePtrList<ObjectLiteral::Property>* NewObjectPropertyList(
-      int size) const {
-    return zone()->New<ZonePtrList<ObjectLiteral::Property>>(size, zone());
-  }
   V8_INLINE ZonePtrList<ClassLiteral::Property>* NewClassPropertyList(
       int size) const {
     return zone()->New<ZonePtrList<ClassLiteral::Property>>(size, zone());
@@ -848,9 +848,6 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
       int size) const {
     return zone()->New<ZonePtrList<ClassLiteral::StaticElement>>(size, zone());
   }
-  V8_INLINE ZonePtrList<Statement>* NewStatementList(int size) const {
-    return zone()->New<ZonePtrList<Statement>>(size, zone());
-  }
 
   Expression* NewV8Intrinsic(const AstRawString* name,
                              const ScopedPtrList<Expression>& args, int pos);
@@ -896,6 +893,8 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
     }
   }
 
+  void ReindexArrowFunctionFormalParameters(ParserFormalParameters* parameters);
+  void ReindexComputedMemberName(Expression* computed_name);
   void DeclareArrowFunctionFormalParameters(
       ParserFormalParameters* parameters, Expression* params,
       const Scanner::Location& params_loc);
@@ -962,6 +961,33 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
                               zone()->New<CaseClauseSourceRanges>(body_range));
   }
 
+  V8_INLINE void AppendConditionalChainSourceRange(ConditionalChain* node,
+                                                   const SourceRange& range) {
+    if (source_range_map_ == nullptr) return;
+    ConditionalChainSourceRanges* ranges =
+        static_cast<ConditionalChainSourceRanges*>(
+            source_range_map_->Find(node));
+    if (ranges == nullptr) {
+      source_range_map_->Insert(
+          node, zone()->New<ConditionalChainSourceRanges>(zone()));
+    }
+    ranges = static_cast<ConditionalChainSourceRanges*>(
+        source_range_map_->Find(node));
+    if (ranges == nullptr) return;
+    ranges->AddThenRanges(range);
+    DCHECK_EQ(node->conditional_chain_length(), ranges->RangeCount());
+  }
+
+  V8_INLINE void AppendConditionalChainElseSourceRange(
+      ConditionalChain* node, const SourceRange& range) {
+    if (source_range_map_ == nullptr) return;
+    ConditionalChainSourceRanges* ranges =
+        static_cast<ConditionalChainSourceRanges*>(
+            source_range_map_->Find(node));
+    if (ranges == nullptr) return;
+    ranges->AddElseRange(range);
+  }
+
   V8_INLINE void RecordConditionalSourceRange(Expression* node,
                                               const SourceRange& then_range,
                                               const SourceRange& else_range) {
@@ -1083,12 +1109,10 @@ class V8_EXPORT_PRIVATE Parser : public NON_EXPORTED_BASE(ParserBase<Parser>) {
 
   LocalIsolate* local_isolate_;
   ParseInfo* info_;
-  Handle<Script> script_;
   Scanner scanner_;
   Zone preparser_zone_;
   PreParser* reusable_preparser_;
   Mode mode_;
-  bool overall_parse_is_parked_ = false;
 
   MaybeHandle<FixedArray> maybe_wrapped_arguments_;
 
