@@ -37,8 +37,7 @@ const {
 
 let workerStdio;
 function lazyWorkerStdio() {
-  if (!workerStdio) workerStdio = createWorkerStdio();
-  return workerStdio;
+  return workerStdio ??= createWorkerStdio();
 }
 
 function getStdout() { return lazyWorkerStdio().stdout; }
