@@ -471,7 +471,7 @@ async function fsCall(fn, handle, ...args) {
 
 function checkAborted(signal) {
   if (signal?.aborted)
-    throw new AbortError(undefined, { cause: signal?.reason });
+    throw new AbortError(undefined, { cause: signal.reason });
 }
 
 async function writeFileHandle(filehandle, data, signal, encoding) {
@@ -620,8 +620,8 @@ async function cp(src, dest, options) {
 async function copyFile(src, dest, mode) {
   return await PromisePrototypeThen(
     binding.copyFile(
-      getValidatedPath(src, 'src', { expectFile: true, syscall: 'cp' }),
-      getValidatedPath(dest, 'dest', { expectFile: true, syscall: 'cp' }),
+      getValidatedPath(src, 'src'),
+      getValidatedPath(dest, 'dest'),
       mode,
       kUsePromises,
     ),
@@ -633,7 +633,7 @@ async function copyFile(src, dest, mode) {
 // Note that unlike fs.open() which uses numeric file descriptors,
 // fsPromises.open() uses the fs.FileHandle class.
 async function open(path, flags, mode) {
-  path = getValidatedPath(path, 'path', { expectFile: true, syscall: 'open' });
+  path = getValidatedPath(path);
   const flagsNumber = stringToFlags(flags);
   mode = parseFileMode(mode, 'mode', 0o666);
   return new FileHandle(await PromisePrototypeThen(
@@ -944,6 +944,10 @@ async function readdirRecursive(originalPath, options) {
 
 async function readdir(path, options) {
   options = getOptions(options);
+
+  // Make shallow copy to prevent mutating options from affecting results
+  options = copyObject(options);
+
   path = getValidatedPath(path);
   if (options.recursive) {
     return readdirRecursive(path, options);
