@@ -4,6 +4,7 @@ const {
   ArrayPrototypePush,
   FunctionPrototypeBind,
   PromiseResolve,
+  PromiseWithResolvers,
   SafeMap,
   SafePromiseAllReturnVoid,
 } = primordials;
@@ -26,7 +27,6 @@ const {
   shouldColorizeTestFiles,
 } = require('internal/test_runner/utils');
 const { queueMicrotask } = require('internal/process/task_queues');
-const { createDeferredPromise } = require('internal/util');
 const { bigint: hrtime } = process.hrtime;
 const resolvedPromise = PromiseResolve();
 const testResources = new SafeMap();
@@ -35,11 +35,11 @@ let globalRoot;
 testResources.set(reporterScope.asyncId(), reporterScope);
 
 function createTestTree(rootTestOptions, globalOptions) {
-  const buildPhaseDeferred = createDeferredPromise();
+  const buildPhaseDeferred = PromiseWithResolvers();
   const isFilteringByName = globalOptions.testNamePatterns ||
                             globalOptions.testSkipPatterns;
-  const isFilteringByOnly = globalOptions.only || (globalOptions.isTestRunner &&
-                                                   globalOptions.isolation === 'none');
+  const isFilteringByOnly = (globalOptions.isolation === 'process' || process.env.NODE_TEST_CONTEXT) ?
+    globalOptions.only : true;
   const harness = {
     __proto__: null,
     buildPromise: buildPhaseDeferred.promise,
@@ -52,7 +52,7 @@ function createTestTree(rootTestOptions, globalOptions) {
     resetCounters() {
       harness.counters = {
         __proto__: null,
-        all: 0,
+        tests: 0,
         failed: 0,
         passed: 0,
         cancelled: 0,
@@ -62,6 +62,7 @@ function createTestTree(rootTestOptions, globalOptions) {
         suites: 0,
       };
     },
+    success: true,
     counters: null,
     shouldColorizeTestFiles: shouldColorizeTestFiles(globalOptions.destinations),
     teardown: null,
@@ -115,7 +116,7 @@ function createProcessEventHandler(eventName, rootTest) {
         const name = test.hookType ? `Test hook "${test.hookType}"` : `Test "${test.name}"`;
         let locInfo = '';
         if (test.loc) {
-          const relPath = relative(process.cwd(), test.loc.file);
+          const relPath = relative(rootTest.config.cwd, test.loc.file);
           locInfo = ` at ${relPath}:${test.loc.line}:${test.loc.column}`;
         }
 
@@ -130,6 +131,7 @@ function createProcessEventHandler(eventName, rootTest) {
       }
 
       rootTest.diagnostic(msg);
+      rootTest.harness.success = false;
       process.exitCode = kGenericUserError;
       return;
     }
@@ -152,6 +154,7 @@ function configureCoverage(rootTest, globalOptions) {
     const msg = `Warning: Code coverage could not be enabled. ${err}`;
 
     rootTest.diagnostic(msg);
+    rootTest.harness.success = false;
     process.exitCode = kGenericUserError;
   }
 }
@@ -167,6 +170,7 @@ function collectCoverage(rootTest, coverage) {
     summary = coverage.summary();
   } catch (err) {
     rootTest.diagnostic(`Warning: Could not report code coverage. ${err}`);
+    rootTest.harness.success = false;
     process.exitCode = kGenericUserError;
   }
 
@@ -174,6 +178,7 @@ function collectCoverage(rootTest, coverage) {
     coverage.cleanup();
   } catch (err) {
     rootTest.diagnostic(`Warning: Could not clean up code coverage. ${err}`);
+    rootTest.harness.success = false;
     process.exitCode = kGenericUserError;
   }
 
@@ -248,14 +253,17 @@ function lazyBootstrapRoot() {
   if (!globalRoot) {
     // This is where the test runner is bootstrapped when node:test is used
     // without the --test flag or the run() API.
+    const entryFile = process.argv?.[1];
     const rootTestOptions = {
       __proto__: null,
-      entryFile: process.argv?.[1],
+      entryFile,
+      loc: entryFile ? [1, 1, entryFile] : undefined,
     };
     const globalOptions = parseCommandLine();
+    globalOptions.cwd = process.cwd();
     createTestTree(rootTestOptions, globalOptions);
-    globalRoot.reporter.on('test:fail', (data) => {
-      if (data.todo === undefined || data.todo === false) {
+    globalRoot.reporter.on('test:summary', (data) => {
+      if (!data.success) {
         process.exitCode = kGenericUserError;
       }
     });
