@@ -2,30 +2,33 @@
 const {
   ArrayPrototypePush,
   ArrayPrototypePushApply,
-  ArrayPrototypeReduce,
   ArrayPrototypeShift,
   ArrayPrototypeSlice,
   ArrayPrototypeSome,
   ArrayPrototypeSplice,
   ArrayPrototypeUnshift,
+  ArrayPrototypeUnshiftApply,
   FunctionPrototype,
   MathMax,
   Number,
   NumberPrototypeToFixed,
   ObjectDefineProperty,
   ObjectSeal,
+  Promise,
   PromisePrototypeThen,
   PromiseResolve,
+  PromiseWithResolvers,
   ReflectApply,
   RegExpPrototypeExec,
   SafeMap,
   SafePromiseAll,
-  SafePromisePrototypeFinally,
+  SafePromiseAllReturnVoid,
   SafePromiseRace,
   SafeSet,
   StringPrototypeStartsWith,
   StringPrototypeTrim,
   Symbol,
+  SymbolDispose,
 } = primordials;
 const { getCallerLocation } = internalBinding('util');
 const { exitCodes: { kGenericUserError } } = internalBinding('errors');
@@ -46,12 +49,11 @@ const {
   createDeferredCallback,
   countCompletedTest,
   isTestFailureError,
+  reporterScope,
 } = require('internal/test_runner/utils');
 const {
-  createDeferredPromise,
   kEmptyObject,
   once: runOnce,
-  SymbolDispose,
 } = require('internal/util');
 const { isPromise } = require('internal/util/types');
 const {
@@ -100,6 +102,7 @@ function lazyAssertObject(harness) {
   if (assertObj === undefined) {
     assertObj = new SafeMap();
     const assert = require('assert');
+    const { SnapshotManager } = require('internal/test_runner/snapshot');
     const methodsToCopy = [
       'deepEqual',
       'deepStrictEqual',
@@ -114,6 +117,7 @@ function lazyAssertObject(harness) {
       'notDeepStrictEqual',
       'notEqual',
       'notStrictEqual',
+      'partialDeepStrictEqual',
       'rejects',
       'strictEqual',
       'throws',
@@ -122,18 +126,14 @@ function lazyAssertObject(harness) {
       assertObj.set(methodsToCopy[i], assert[methodsToCopy[i]]);
     }
 
-    const { getOptionValue } = require('internal/options');
-    if (getOptionValue('--experimental-test-snapshots')) {
-      const { SnapshotManager } = require('internal/test_runner/snapshot');
-      harness.snapshotManager = new SnapshotManager(harness.config.updateSnapshots);
-      assertObj.set('snapshot', harness.snapshotManager.createAssert());
-    }
+    harness.snapshotManager = new SnapshotManager(harness.config.updateSnapshots);
+    assertObj.set('snapshot', harness.snapshotManager.createAssert());
   }
   return assertObj;
 }
 
 function stopTest(timeout, signal) {
-  const deferred = createDeferredPromise();
+  const deferred = PromiseWithResolvers();
   const abortListener = addAbortListener(signal, deferred.resolve);
   let timer;
   let disposeFunction;
@@ -556,7 +556,7 @@ class Test extends AsyncResource {
         const map = lazyFindSourceMap(this.loc.file);
         const entry = map?.findEntry(this.loc.line - 1, this.loc.column - 1);
 
-        if (entry !== undefined) {
+        if (entry?.originalSource !== undefined) {
           this.loc.line = entry.originalLine + 1;
           this.loc.column = entry.originalColumn + 1;
           this.loc.file = entry.originalSource;
@@ -653,7 +653,7 @@ class Test extends AsyncResource {
       // - Only called if there are results to report in the correct order.
       // - Guaranteed to only be called a maximum of once per call to
       //   processReadySubtestRange().
-      canSend = canSend || this.isClearToSend();
+      canSend ||= this.isClearToSend();
 
       if (!canSend) {
         return;
@@ -732,9 +732,9 @@ class Test extends AsyncResource {
 
   computeInheritedHooks() {
     if (this.parent.hooks.beforeEach.length > 0) {
-      ArrayPrototypeUnshift(
+      ArrayPrototypeUnshiftApply(
         this.hooks.beforeEach,
-        ...ArrayPrototypeSlice(this.parent.hooks.beforeEach),
+        ArrayPrototypeSlice(this.parent.hooks.beforeEach),
       );
     }
 
@@ -818,7 +818,7 @@ class Test extends AsyncResource {
     // pending for later execution.
     this.reporter.enqueue(this.nesting, this.loc, this.name);
     if (this.root.harness.buildPromise || !this.parent.hasConcurrency()) {
-      const deferred = createDeferredPromise();
+      const deferred = PromiseWithResolvers();
 
       deferred.test = this;
       this.parent.addPendingSubtest(deferred);
@@ -844,13 +844,14 @@ class Test extends AsyncResource {
   async runHook(hook, args) {
     validateOneOf(hook, 'hook name', kHookNames);
     try {
-      await ArrayPrototypeReduce(this.hooks[hook], async (prev, hook) => {
-        await prev;
+      const hooks = this.hooks[hook];
+      for (let i = 0; i < hooks.length; ++i) {
+        const hook = hooks[i];
         await hook.run(args);
         if (hook.error) {
           throw hook.error;
         }
-      }, PromiseResolve());
+      }
     } catch (err) {
       const error = new ERR_TEST_FAILURE(`failed running ${hook} hook`, kHookFailure);
       error.cause = isTestFailureError(err) ? err.cause : err;
@@ -973,10 +974,25 @@ class Test extends AsyncResource {
       // any remaining ref'ed handles, then do that now. It is theoretically
       // possible that a ref'ed handle could asynchronously create more tests,
       // but the user opted into this behavior.
-      this.reporter.once('close', () => {
-        process.exit();
-      });
+      const promises = [];
+
+      for (let i = 0; i < reporterScope.reporters.length; i++) {
+        const { destination } = reporterScope.reporters[i];
+
+        ArrayPrototypePush(promises, new Promise((resolve) => {
+          destination.on('unpipe', () => {
+            if (!destination.closed && typeof destination.close === 'function') {
+              destination.close(resolve);
+            } else {
+              resolve();
+            }
+          });
+        }));
+      }
+
       this.harness.teardown();
+      await SafePromiseAllReturnVoid(promises);
+      process.exit();
     }
   }
 
@@ -1043,14 +1059,15 @@ class Test extends AsyncResource {
         reporter.diagnostic(nesting, loc, diagnostics[i]);
       }
 
-      reporter.diagnostic(nesting, loc, `tests ${harness.counters.all}`);
+      const duration = this.duration();
+      reporter.diagnostic(nesting, loc, `tests ${harness.counters.tests}`);
       reporter.diagnostic(nesting, loc, `suites ${harness.counters.suites}`);
       reporter.diagnostic(nesting, loc, `pass ${harness.counters.passed}`);
       reporter.diagnostic(nesting, loc, `fail ${harness.counters.failed}`);
       reporter.diagnostic(nesting, loc, `cancelled ${harness.counters.cancelled}`);
       reporter.diagnostic(nesting, loc, `skipped ${harness.counters.skipped}`);
       reporter.diagnostic(nesting, loc, `todo ${harness.counters.todo}`);
-      reporter.diagnostic(nesting, loc, `duration_ms ${this.duration()}`);
+      reporter.diagnostic(nesting, loc, `duration_ms ${duration}`);
 
       if (coverage) {
         const coverages = [
@@ -1067,6 +1084,7 @@ class Test extends AsyncResource {
         for (let i = 0; i < coverages.length; i++) {
           const { threshold, actual, name } = coverages[i];
           if (actual < threshold) {
+            harness.success = false;
             process.exitCode = kGenericUserError;
             reporter.diagnostic(nesting, loc, `Error: ${NumberPrototypeToFixed(actual, 2)}% ${name} coverage does not meet threshold of ${threshold}%.`);
           }
@@ -1075,6 +1093,10 @@ class Test extends AsyncResource {
         reporter.coverage(nesting, loc, coverage);
       }
 
+      reporter.summary(
+        nesting, loc?.file, harness.success, harness.counters, duration,
+      );
+
       if (harness.watching) {
         this.reported = false;
         harness.resetCounters();
@@ -1236,27 +1258,20 @@ class Suite extends Test {
       this.skipped = false;
     }
 
+    this.buildSuite = this.createBuild();
+    this.fn = noop;
+  }
+
+  async createBuild() {
     try {
       const { ctx, args } = this.getRunArgs();
       const runArgs = [this.fn, ctx];
       ArrayPrototypePushApply(runArgs, args);
-      this.buildSuite = SafePromisePrototypeFinally(
-        PromisePrototypeThen(
-          PromiseResolve(ReflectApply(this.runInAsyncScope, this, runArgs)),
-          undefined,
-          (err) => {
-            this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
-          }),
-        () => this.postBuild(),
-      );
+      await ReflectApply(this.runInAsyncScope, this, runArgs);
     } catch (err) {
       this.fail(new ERR_TEST_FAILURE(err, kTestCodeFailure));
-      this.postBuild();
     }
-    this.fn = noop;
-  }
 
-  postBuild() {
     this.buildPhaseFinished = true;
   }
 
