@@ -107,7 +107,9 @@ function getDeprecationWarningEmitter(
   return function() {
     if (!warned && shouldEmitWarning()) {
       warned = true;
-      if (code !== undefined) {
+      if (code === 'ExperimentalWarning') {
+        process.emitWarning(msg, code, deprecated);
+      } else if (code !== undefined) {
         if (!codesWarned.has(code)) {
           const emitWarning = useEmitSync ?
             require('internal/process/warning').emitWarningSync :
@@ -266,14 +268,20 @@ function slowCases(enc) {
   }
 }
 
-function emitExperimentalWarning(feature, messagePrefix) {
+/**
+ * @param {string} feature Feature name used in the warning message
+ * @param {string} messagePrefix Prefix of the warning message
+ * @param {string} code See documentation of process.emitWarning
+ * @param {string} ctor See documentation of process.emitWarning
+ */
+function emitExperimentalWarning(feature, messagePrefix, code, ctor) {
   if (experimentalWarnings.has(feature)) return;
   experimentalWarnings.add(feature);
   let msg = `${feature} is an experimental feature and might change at any time`;
   if (messagePrefix) {
     msg = messagePrefix + msg;
   }
-  process.emitWarning(msg, 'ExperimentalWarning');
+  process.emitWarning(msg, 'ExperimentalWarning', code, ctor);
 }
 
 function filterDuplicateStrings(items, low) {
