@@ -28,6 +28,7 @@ const {
 const { compare } = internalBinding('buffer');
 const assert = require('internal/assert');
 const { isURL } = require('internal/url');
+const { isError } = require('internal/util');
 const types = require('internal/util/types');
 const {
   isAnyArrayBuffer,
@@ -235,10 +236,10 @@ function innerDeepEqual(val1, val2, strict, memos) {
     if (!isAnyArrayBuffer(val2) || !areEqualArrayBuffers(val1, val2)) {
       return false;
     }
-  } else if (isNativeError(val1) || val1 instanceof Error) {
+  } else if (isError(val1)) {
     // Do not compare the stack as it might differ even though the error itself
     // is otherwise identical.
-    if (!isNativeError(val2) && !(val2 instanceof Error)) {
+    if (!isError(val2)) {
       return false;
     }
 
@@ -252,7 +253,9 @@ function innerDeepEqual(val1, val2, strict, memos) {
         (name1Enumerable !== ObjectPrototypePropertyIsEnumerable(val2, 'name') ||
           (!name1Enumerable && val1.name !== val2.name)) ||
         (cause1Enumerable !== ObjectPrototypePropertyIsEnumerable(val2, 'cause') ||
-          (!cause1Enumerable && !innerDeepEqual(val1.cause, val2.cause, strict, memos))) ||
+          (!cause1Enumerable && (
+            ObjectPrototypeHasOwnProperty(val1, 'cause') !== ObjectPrototypeHasOwnProperty(val2, 'cause') ||
+            !innerDeepEqual(val1.cause, val2.cause, strict, memos)))) ||
         (errors1Enumerable !== ObjectPrototypePropertyIsEnumerable(val2, 'errors') ||
          (!errors1Enumerable && !innerDeepEqual(val1.errors, val2.errors, strict, memos)))) {
       return false;
