@@ -130,7 +130,7 @@ const { shouldColorize } = require('internal/util/colors');
 const CJSModule = require('internal/modules/cjs/loader').Module;
 let _builtinLibs = ArrayPrototypeFilter(
   CJSModule.builtinModules,
-  (e) => !StringPrototypeStartsWith(e, '_'),
+  (e) => e[0] !== '_' && !StringPrototypeStartsWith(e, 'node:'),
 );
 const nodeSchemeBuiltinLibs = ArrayPrototypeMap(
   _builtinLibs, (lib) => `node:${lib}`);
@@ -286,10 +286,9 @@ function REPLServer(prompt,
 
   if (!options.input && !options.output) {
     // Legacy API, passing a 'stream'/'socket' option.
-    if (!stream) {
-      // Use stdin and stdout as the default streams if none were given.
-      stream = process;
-    }
+    // Use stdin and stdout as the default streams if none were given.
+    stream ||= process;
+
     // We're given a duplex readable/writable Stream, like a `net.Socket`
     // or a custom object with 2 streams, or the `process` object.
     options.input = stream.stdin || stream;
@@ -398,7 +397,7 @@ function REPLServer(prompt,
                                   `${sep}(.*)${sep}(.*)${sep}(.*)${sep}(.*)` +
                                   `${sep}(.*)$`);
 
-  eval_ = eval_ || defaultEval;
+  eval_ ||= defaultEval;
 
   const self = this;
 
@@ -879,7 +878,7 @@ function REPLServer(prompt,
 
   self.on('line', function onLine(cmd) {
     debug('line %j', cmd);
-    cmd = cmd || '';
+    cmd ||= '';
     sawSIGINT = false;
 
     if (self.editorMode) {
@@ -908,8 +907,8 @@ function REPLServer(prompt,
           StringPrototypeCharAt(trimmedCmd, 1) !== '.' &&
           NumberIsNaN(NumberParseFloat(trimmedCmd))) {
         const matches = RegExpPrototypeExec(/^\.([^\s]+)\s*(.*)$/, trimmedCmd);
-        const keyword = matches && matches[1];
-        const rest = matches && matches[2];
+        const keyword = matches?.[1];
+        const rest = matches?.[2];
         if (ReflectApply(_parseREPLKeyword, self, [keyword, rest]) === true) {
           return;
         }
@@ -1007,7 +1006,7 @@ function REPLServer(prompt,
   // Wrap readline tty to enable editor mode and pausing.
   const ttyWrite = FunctionPrototypeBind(self._ttyWrite, self);
   self._ttyWrite = (d, key) => {
-    key = key || {};
+    key ||= {};
     if (paused && !(self.breakEvalOnSigint && key.ctrl && key.name === 'c')) {
       ArrayPrototypePush(pausedBuffer,
                          ['key', [d, key], self.isCompletionEnabled]);
@@ -1253,7 +1252,7 @@ function filteredOwnPropertyNames(obj) {
   let isObjectPrototype = false;
   if (ObjectGetPrototypeOf(obj) === null) {
     const ctorDescriptor = ObjectGetOwnPropertyDescriptor(obj, 'constructor');
-    if (ctorDescriptor && ctorDescriptor.value) {
+    if (ctorDescriptor?.value) {
       const ctorProto = ObjectGetPrototypeOf(ctorDescriptor.value);
       isObjectPrototype = ctorProto && ObjectGetPrototypeOf(ctorProto) === obj;
     }
@@ -1528,7 +1527,7 @@ function complete(line, callback) {
         let p;
         if ((typeof obj === 'object' && obj !== null) ||
             typeof obj === 'function') {
-          memberGroups.push(filteredOwnPropertyNames(obj));
+          ArrayPrototypePush(memberGroups, filteredOwnPropertyNames(obj));
           p = ObjectGetPrototypeOf(obj);
         } else {
           p = obj.constructor ? obj.constructor.prototype : null;
@@ -1536,7 +1535,7 @@ function complete(line, callback) {
         // Circular refs possible? Let's guard against that.
         let sentinel = 5;
         while (p !== null && sentinel-- !== 0) {
-          memberGroups.push(filteredOwnPropertyNames(p));
+          ArrayPrototypePush(memberGroups, filteredOwnPropertyNames(p));
           p = ObjectGetPrototypeOf(p);
         }
       } catch {
@@ -1553,9 +1552,7 @@ function complete(line, callback) {
                              ArrayPrototypeMap(group,
                                                (member) => `${expr}${member}`));
         });
-        if (filter) {
-          filter = `${expr}${filter}`;
-        }
+        filter &&= `${expr}${filter}`;
       }
 
       completionGroupsLoaded();
@@ -1645,8 +1642,8 @@ REPLServer.prototype.defineCommand = function(keyword, cmd) {
 // sufficient anymore.
 function _memory(cmd) {
   const self = this;
-  self.lines = self.lines || [];
-  self.lines.level = self.lines.level || [];
+  self.lines ||= [];
+  self.lines.level ||= [];
 
   // Save the line so I can do magic later
   if (cmd) {
