@@ -78,6 +78,18 @@ const filename = 'foo';
     },
     ));
 }
+{
+  // Reassigning `.path` property should not trigger a warning
+  const dirent = getDirent(
+    tmpdir.path,
+    filename,
+    UV_DIRENT_UNKNOWN,
+  );
+  assert.strictEqual(dirent.name, filename);
+  dirent.path = 'some other value';
+  assert.strictEqual(dirent.parentPath, tmpdir.path);
+  assert.strictEqual(dirent.path, 'some other value');
+}
 {
   // string + Buffer
   const filenameBuffer = Buffer.from(filename);
